#!/usr/bin/perl -w
use strict;
#use diagnostics;

# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is the Bugzilla Bug Tracking System.
#
# The Initial Developer of the Original Code is Netscape Communications
# Corporation. Portions created by Netscape are
# Copyright (C) 1998 Netscape Communications Corporation. All
# Rights Reserved.
#
# Contributor(s): Terry Weissman <terry@mozilla.org>
#                 Dan Mosedale <dmose@mozilla.org>
#                 Joe Robins <jmrobins@tgix.com>
#                 Dave Miller <justdave@syndicomm.com>
#                 Frederick Dean <software@fdd.com>

package cgi;

# Decode the escaping of a URL
sub urlUnquote {
   my ($tounquote) = (@_);
   $tounquote =~ tr/+/ /;       # pluses become spaces
   $tounquote =~ s/%([0-9a-fA-F]{2})/pack("c",hex($1))/ge;
   return $tounquote;
}

# Quotify a string, suitable for putting into a URL.
sub urlQuote {
   my($toencode) = (@_);
   $toencode=~s/([^a-zA-Z0-9_\-.])/uc sprintf("%%%02x",ord($1))/eg;
   return $toencode;
}

sub Fatal {
   my ($error) = (@_);

   common::content_type("text/html");
   print("  <br><font color=red size=+2>Fatal Error</font><br>\n".
         "<font color=red>" . htmlQuote($error) . "</font>\n" .
         "</td></tr></table>");
   exit(1);
}

our %form;  # This is the has of form data submitted
           # The keys are the "name=" of the HTML form

# useful for debugging.  Prints %form 
sub PrintFormHash {
   print("<hr><font size=+1>Form hash</font>\n<table>\n");
   my $rownum = 0;
   for my $key (sort(keys(%form))) {
      (my $value = $form{$key}) =~ s/\n/<br>\n/m;
      my $color = ($rownum++ & 1) ? "ffffff" : "dddddd";
      print("<tr bgcolor=$color><td>$key</td><td><font color=blue>$value</font></td></tr>\n"); 
   }
   print("</table>\n");
}

# Here we parse the HTML-form data from apache into a global public hash $cgi::form{} 
sub MakeFormHash {
   my $queryString = "";
   if(defined($ENV{"REQUEST_METHOD"}) and $ENV{"REQUEST_METHOD"} eq "POST") {
      read(STDIN,$queryString,$ENV{"CONTENT_LENGTH"});
   } elsif(defined($ENV{"QUERY_STRING"})) {
      $queryString = $ENV{"QUERY_STRING"};
   }
   # $queryString looks like pw=passwd&name=rick+dean
   my @pairs = split('&',$queryString);
   for my $pair (@pairs) {
      if($pair =~ /([^=]*)=(.*)/) {
         $form{$1} = urlUnquote($2);
      };
   }
}

our %cookie;
our $cookielessFlag = 0;

# useful for debugging.  Prints %form 
sub PrintCookieHash {
   print("<hr><font size=+1>Cookies hash</font><br>\n<font color=blue>\n");
   for my $key (sort(keys(%cookie))) {
      print("$key => $cookie{$key}<br>\n"); 
   }
   print("</font>\n");
}

sub GetPathAfterCgi {
   my $path = $ENV{'PATH_INFO'};
   if(! defined $path) {
      $path = $ENV{'REQUEST_URI'};
      return "" if ! defined($path);
      my $script_name = $ENV{'SCRIPT_NAME'};
      if($script_name) {
         $script_name =~ s/(\W)/\\$1/g;  # escape for regex
         $path =~ s/^$script_name//i;  # delete script name prefix
      };
      $path =~ s/\?.*$//;  # trim question mark and following
   }
   $path = urlUnquote($path);
   # For people who won't use cookies, we pass it as a bogus first directory.
   # If we already have a session cookie, forget about the URI one
#print("cookielessFlag $cookielessFlag  <font color=green>$cookie{'dcvs_session'}</font><br>\n");
   if($path =~ s|/-([\w-]{22})-/|/| && !$cookie{'dcvs_session'}) {
      $cookie{'dcvs_session'} = $1;
      $cookielessFlag = 1;   # need to know when we make links
   }
   return $path;
}

# useful for debugging
sub PrintEnv {
   print("<hr><font size=+1>Environment</font><br>\n");
   for my $key (sort(keys(%ENV))) {
      print("$key => $ENV{$key}<br>\n"); 
   }
}

# This makes a new link being smart about the state we want to keep.
# The args are a list of "param=value".  
# The value should not be URL encoded because this function will.
sub Uri 
{
   my ($filename,@params) = @_;

   my %answer;
   my @keepers = qw{ pw sort branch cols latest showRm diskPath dbPath};
   # get all the keeper from the current page
   for my $keeper (@keepers) {  # for every keeper
      if(defined($form{$keeper})) {
         $answer{$keeper} = $form{$keeper};
      };
   };
   # incorporate all the new parameters
   for my $parampair (@params) {   # for every new parameter
      next if(!defined($parampair));
      my($param,$value) = split("=",$parampair,2);
      $answer{$param} = $value if $value ne "";
   };
   my $answer = "";
   for my $key (keys(%answer)) {  # for every answer param
      $answer .= "$key=" . urlQuote($answer{$key}) . '&'; # append
   }
   chop($answer); # kill trailing '&'
   $answer = "?$answer" if $answer ne "";
   my $cookie = "";
   if($cookielessFlag && $cookie{'dcvs_session'}) {
      $cookie = "/-$cookie{'dcvs_session'}-";
   };
   $filename = GetPathAfterCgi() if $filename eq "";
   return("$ENV{'SCRIPT_NAME'}$cookie$filename$answer");
}
sub Link
{
   return("<a href=\"". Uri(@_) ."\">");
}


# returns a string converted to html
# double spaces and newlines are not processed
sub htmlQuote {
   my ($html) = (@_);
   return "&nbsp;" if !defined($html) || $html eq "";
   $html =~ s/\&/\&amp;/g;
   $html =~ s/</\&lt;/g;
   $html =~ s/>/\&gt;/g;
   $html =~ s/\n/\n<br>/g;
   return $html;
}
sub htmlQuoteForPre {
   my ($html) = (@_);
   return "&nbsp;" if !defined($html) || $html eq "";
   $html =~ s/\&/\&amp;/g;
   $html =~ s/</\&lt;/g;
   $html =~ s/>/\&gt;/g;
   return $html;
}
# Create a new has of everythin html quoted
sub htmlQuoteHash {
   my ($orig) = (@_);

   my %quoted;
   for my $key (keys(%$orig)) {
      $quoted{$key} = htmlQuote($$orig{$key});
   };
   return \%quoted;
}

our %keyword;  # This is the hash which stores all the replacements for the templates
sub PrintKeywordHash
{
   print("<hr>\n<font size=+1>Keyword hash</font>\n<table>\n");
   for my $key (sort(keys(%keyword))) {
      print("<tr><td bgcolor=dddddd>$key</td><td>$keyword{$key}</td></tr>\n");
   }
   print("</table>\n");
}

sub expanded_keyword {
    my($keyword) = @_;
    my $answer = $keyword{$keyword};
    if(!defined($answer)) {
       if(-r "template/$keyword-template.html") {
          $answer = expand_template_keywords("template/$keyword-template.html"); 
       } else {
          $answer = "<font color=orange>{{$keyword}}</font>";
       };
    };
    return $answer;
}

# This function uses the hash above. 
sub expand_template_keywords {
    my($templateFilename) = @_;
    local(*FILE);
    open(FILE, "<$templateFilename") or die "Couldn't open $templateFilename: $!\n";
    read(FILE,my $answer,100000);
    close FILE;
    #$answer =~ s/{{(.*?)}}/$keyword{$1}/g;
    $answer =~ s/{{(.*?)}}/expanded_keyword($1)/eg;
    return($answer);
}

#
# make_selection_widget: creates an HTML selection widget from a list of text strings.
# $groupname is the name of the setting (form value) that this widget will control
# $src is the list of options
# you can specify a $default value which is either a string or a regex pattern to match to
#    identify the default value
# $capitalize lets you optionally capitalize the option strings; the default is the value
#    of Param("capitalizelists")
# $multiple is 1 if several options are selectable (default), 0 otherwise.
# $size is used for lists to control how many items are shown. The default is 7. A list of
#    size 1 becomes a popup menu.
# $preferLists is 1 if selection lists should be used in favor of radio buttons and
#    checkboxes, and 0 otherwise. The default is the value of Param("preferlists").
#
# The actual widget generated depends on the parameter settings:
# 
#        MULTIPLE     PREFERLISTS    SIZE     RESULT
#       0 (single)        0           =1      Popup Menu (normal for list of size 1)
#       0 (single)        0           >1      Radio buttons
#       0 (single)        1           =1      Popup Menu (normal for list of size 1)
#       0 (single)        1           n>1     List of size n, single selection
#       1 (multi)         0           n/a     Check boxes; size ignored
#       1 (multi)         1           n/a     List of size n, multiple selection, of size n
#
sub make_selection_widget {
    my ($groupname,$src,$default,$isregexp,$multiple, $size, $capitalize, $preferLists) = (@_);
    my $last = "";
    my $popup = "";
    my $found = 0;
    my $displaytext = "";
    $groupname = "" if !defined $groupname;
    $default = "" if !defined $default;
    #$capitalize = Param("capitalizelists") if !defined $capitalize;
    $multiple = 1 if !defined $multiple;
    #$preferLists = Param("preferlists") if !defined $preferLists;
    $size = 7 if !defined $size;
    my $type = "LIST";
    if (!$preferLists) {
        if ($multiple) {
            $type = "CHECKBOX";
        } else {
            if ($size > 1) {
                $type = "RADIO";
            }
        }
    }

    if ($type eq "LIST") {
        $popup .= "<SELECT NAME=\"$groupname\"";
        if ($multiple) {
            $popup .= " MULTIPLE";
        }
        $popup .= " SIZE=$size>\n";
    }
    if ($src) {
        foreach my $item (@$src) {
            if ($item eq "-blank-" || $item ne $last) {
                if ($item eq "-blank-") {
                    $item = "";
                }
                $last = $item;
                $displaytext = $item;
                if ($capitalize) {
                    $displaytext =~ tr/A-Z/a-z/;
                    $displaytext =~ s/^(.?)(.*)/\u$1$2/;
                }   

                if ($isregexp ? $item =~ $default : $default eq $item) {
                    if ($type eq "CHECKBOX") {
                        $popup .= "<INPUT NAME=$groupname type=checkbox VALUE=\"$item\" CHECKED>$displaytext<br>";
                    } elsif ($type eq "RADIO") {
                        $popup .= "<INPUT NAME=$groupname type=radio VALUE=\"$item\" check>$displaytext<br>";
                    } else {
                        $popup .= "<OPTION SELECTED VALUE=\"$item\">$displaytext";
                    }
                    $found = 1;
                } else {
                    if ($type eq "CHECKBOX") {
                        $popup .= "<INPUT NAME=$groupname type=checkbox VALUE=\"$item\">$displaytext<br>";
                    } elsif ($type eq "RADIO") {
                        $popup .= "<INPUT NAME=$groupname type=radio VALUE=\"$item\">$displaytext<br>";
                    } else {
                        $popup .= "<OPTION VALUE=\"$item\">$displaytext";
                    }
                }
            }
        }
    }
    if (!$found && $default ne "") {
        if ($type eq "CHECKBOX") {
            $popup .= "<INPUT NAME=$groupname type=checkbox CHECKED>$default";
        } elsif ($type eq "RADIO") {
            $popup .= "<INPUT NAME=$groupname type=radio checked>$default";
        } else {
            $popup .= "<OPTION SELECTED>$default";
        }
    }
    if ($type eq "LIST") {
        $popup .= "</SELECT>";
    }
    return $popup;
}
=pod

$::CheckOptionValues = 1;

sub make_options {
    my ($src,$default,$isregexp) = (@_);
    my $last = "";
    my $popup = "";
    my $found = 0;
    $default = "" if !defined $default;

    if ($src) {
        foreach my $item (@$src) {
            if ($item eq "-blank-" || $item ne $last) {
                if ($item eq "-blank-") {
                    $item = "";
                }
                $last = $item;
                if ($isregexp ? $item =~ $default : $default eq $item) {
                    $popup .= "<OPTION SELECTED VALUE=\"$item\">$item";
                    $found = 1;
                } else {
                    $popup .= "<OPTION VALUE=\"$item\">$item";
                }
            }
        }
    }
    if (!$found && $default ne "") {
      if ( Param("strictvaluechecks") && $::CheckOptionValues &&
           ($default ne $::dontchange) && ($default ne "-All-") &&
           ($default ne "DUPLICATE") ) {
        print "Possible bug database corruption has been detected.  " .
              "Please send mail to " . Param("maintainer") . " with " .
              "details of what you were doing when this message " . 
              "appeared.  Thank you.\n";
        if (!$src) {
            $src = ["???null???"];
        }
        print "<pre>src = " . value_quote(join(' ', @$src)) . "\n";
        print "default = " . value_quote($default) . "</pre>";
        PutFooter();
#        confess "Gulp.";
        exit 0;
              
      } else {
	$popup .= "<OPTION SELECTED>$default";
      }
    }
    return $popup;
}


sub make_popup {
    my ($name,$src,$default,$listtype,$onchange) = (@_);
    my $popup = "<SELECT NAME=$name";
    if ($listtype > 0) {
        $popup .= " SIZE=5";
        if ($listtype == 2) {
            $popup .= " MULTIPLE";
        }
    }
    if (defined $onchange && $onchange ne "") {
        $popup .= " onchange=$onchange";
    }
    $popup .= ">" . make_options($src, $default,
                                 ($listtype == 2 && $default ne ""));
    $popup .= "</SELECT>";
    return $popup;
}


sub BuildPulldown {
    my ($name, $valuelist, $default) = (@_);

    my $entry = qq{<SELECT NAME="$name">};
    foreach my $i (@$valuelist) {
        my ($tag, $desc) = (@$i);
        my $selectpart = "";
        if ($tag eq $default) {
            $selectpart = " SELECTED";
        }
        if (!defined $desc) {
            $desc = $tag;
        }
        $entry .= qq{<OPTION$selectpart VALUE="$tag">$desc\n};
    }
    $entry .= qq{</SELECT>};
    return $entry;
}

=cut

sub MakeCookieHash {
   if (defined $ENV{"HTTP_COOKIE"}) {
      foreach my $pair (split(/;/, $ENV{"HTTP_COOKIE"})) {
         if ($pair =~ /^\s*([^=]*)=(.*)\s*$/) {
             $cookie{$1} = $2;
         } else {
             $cookie{$pair} = "";
         }
      }
   }
}

1;



1;
