#!/usr/bin/perl -w
use strict;
#use diagnostics;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

require cvsdb;
require cvsweb;

package common_html;

sub linkedPath_keyword {
   my($filename,$inforef) = @_; 

   my $linkedPath;
   my $reconstructed = "";
   $linkedPath .= cgi::Link("/") . "root</a>\n";
   $filename =~ s/^\///;  # remove leading slash
   for my $dir (split('/',$filename)) {
      $reconstructed .= "/$dir";
      $linkedPath .= "/" . cgi::Link("$reconstructed") . " $dir </a> ";
   }
   $cgi::keyword{'linkedPath'} = $linkedPath;
}

sub bugzillifyHtml {
   return $_[0] if $config::bugzillaLink eq "";  # do nothing
   my($html) = (@_);
   $html =~ s/(\d{$config::minNumBugDigits,20})/<a href=\"$config::bugzillaLink$1\">$1<\/a>/go;
   return $html;
}

# The $revref is a complicated hash of revision numbers, each one a hash of info, see cvsdb::GetRevInfoFromDb()
sub HtmlRevInfo 
{
   my($filename,$file_id,$revsref) = @_;
   my($tagHashRevKeyRef,$tagHashNameKeyRef) = cvsdb::GetTagHashForFile($file_id);
   my($branchHashNumKeyRef,$branchHashNameKeyRef) = cvsdb::GetBranchHashForFile($file_id);
   my $branchHeadHashRef = cvsdb::GetBranchHeadHashForFile($file_id);
   my @revs = keys(%$revsref);
   @revs = sort { ($$revsref{$a}->{'timestamp'} <=> $$revsref{$b}->{'timestamp'}) || cvsdb::rev_cmp($a,$b)} @revs; 
   db::getDbNowTime();
   my $numRevs = 0;
   my $out = "";
   while($#revs != -1) {
      my $rev = pop(@revs);
      $out .= "<hr noshade>\n" if $numRevs++;
      $out .= "Revision: ". cgi::Link($filename,"rev=$rev") . "<b>".cgi::htmlQuote($rev). "</b></a>\n";
      $out .= "<i>". cgi::htmlQuote(::time2str($config::timeformat , $$revsref{$rev}->{'timestamp'})). "</i> ";
      $out .= "(" . cvsweb::readableTime($db::now - $$revsref{$rev}->{'timestamp'},1) . " ago)\n" . " ";
      $out .= "<i>by ".cgi::htmlQuote($$revsref{$rev}->{'author'}). "</i><br>\n";
      my $branch = $$branchHashNumKeyRef{$$revsref{$rev}->{'branchnum'}};
      #my $branch = $$revsref{$rev}->{'branchnum'};
      $out .= "Branch: ". cgi::Link($filename,"branch=$branch") ."<b>". cgi::htmlQuote($branch). "</b></a><br>\n" if defined($branch);
      if(my $tags = $$tagHashRevKeyRef{$rev}) {
         $out .= "Tags: ";
         for my $tag (split(/\s+/,$tags)) {
            $out .= cgi::Link($filename,"tag=$tag") . "<b>" . cgi::htmlQuote($tag). "</b></a>,&nbsp; ";
         };
         $out .= "<br>\n";
      };
      if(my $branches = $$revsref{$rev}->{'branches'}) {
         $out .= "Branch point for: ";
         for my $branchnum (split(/\s+/,$branches)) {
            (my $trunknum = $branchnum) =~ s/\.[^\.]$//;
            my $branchname = $$branchHashNumKeyRef{$trunknum} || $branchnum;
            $out .= cgi::Link($filename,"branch=$branchnum") . "<b>" . cgi::htmlQuote($branchname). "</b></a>, ";
         };
         $out .= "<br>\n";
      };
      if(my $branchnames = $$branchHeadHashRef{$rev}) {
         $out .= "Branch head of: ";
         for my $branchname (split(/\s+/,$branchnames)) {
            $out .= cgi::Link($filename,"branch=$branchname") . "<b>" . cgi::htmlQuote($branchname). "</b></a>, ";
         };
         $out .= "<br>\n";
      }
      if($$revsref{$rev}->{'state'} eq "dead") {
         $out .= "<i><b>FILE REMOVED</b></i><br>\n";
      } else {
         my $prev = cvsdb::PreviousRev($rev,\%$revsref);
         $out .= "Diff to previous: ". cgi::Link("$filename.diff","r1=$rev","r2=$prev") . "<b>$prev</b></a><br>\n" if $prev;
      };
      chomp(my $comment = $$revsref{$rev}->{'comment'});
      $comment = bugzillifyHtml(cgi::htmlQuoteForPre($comment));
      $out .= "<pre>$comment</pre>\n\n";
   };
   $cgi::keyword{'numRevs'} = $numRevs;
   return $out;
} 

#   common_html::branchesWidget_keyword("dirent.parent = $file_id");
#   common_html::branchesWidget_keyword("dirent.file_id = $file_id");

sub branchesWidget_keyword
{
   my($where) = @_;

   db::SendSQL("SELECT branchname FROM btime, dirent WHERE btime.file_id = dirent.file_id AND $where GROUP BY branchname");
   my @branches = ("all");
   while(my $branchname = db::FetchOneColumn()) {
      push(@branches,$branchname);
   }
   $cgi::keyword{'branchesWidget'} = cgi::make_selection_widget("branch",\@branches,$cgi::form{'branch'} || "all",0,0,0,0,1);
}

sub url_keywords 
{
   my($path) = @_;

   $cgi::keyword{"rooturl"} = cgi::Link("/");
   $cgi::keyword{"rooturl"} =~ s/.*?"(.*)">/$1/;  # remove the <a href=
   $cgi::keyword{"selfurl"} = cgi::Link($path);    # without the --xx.html stuff
   $cgi::keyword{"selfurl"} =~ s/.*?"(.*)">/$1/;  # remove the <a href=
   if($cgi::keyword{"selfurl"} =~ s/(\?.*)$//) {  # remove the <a href=
      $cgi::keyword{"tailurl"} = $1; 
   } else {
      $cgi::keyword{"tailurl"} = ""; 
   };
   $cgi::keyword{"nowtime"} = ::time2str("%C",db::getDbNowTime(),$config::timezone); 
}

my $head_done_flag = 0;

sub content_type($) 
{
   my($type) = (@_);

   return if $head_done_flag++;
   print("content-type: $type\n\n");
}





1;

