#!/usr/bin/perl -w
use strict;
#use diagnostics;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

require cgi;
require db;

package cvsdb;

#  There is not a lot of checking here.
#  $name does not include the path.
#  $parent is the numeric dirent table file_id. 
#  We update the descendent table.
sub createDir {
   my($name,$parent) = (@_);

   $parent >= 0 or die;
   db::SendSQL("INSERT dirent SET fname = ". db::SqlQuote($name) .", type = 'dir', parent = $parent");
   db::SendSQL("select LAST_INSERT_ID()");
   my $file_id = db::FetchOneColumn();
   my $ancestor = $file_id;  # list self first
   my $fullyQualified = "";  # qualified with an absolute path.  (we build it backwards)
   # file_id=0 is the root directory.  It does not get descendent entries.
   while($ancestor > 0) {  
      db::SendSQL("REPLACE descendent SET parent = $ancestor, child = $file_id");
      db::SendSQL("SELECT parent,fname FROM dirent WHERE file_id = $ancestor");
      ($ancestor,my $fname) = db::FetchSQLData();
      $fullyQualified = "/$fname$fullyQualified";
   };
=trash   
   # This is an alternate method which doesn't fetch the names
   db::SendSQL("INSERT descendent SET parent = $parent_id, child = $dir_id");
   db::SendSQL("CREATE TEMPORARY TABLE ancestors ".
               "SELECT $dir_id AS child, parent FROM descendent ".
               "WHERE child = $parent_id ");
   db::SendSQL("INSERT descendent SELECT parent, child FROM ancestors");  # Ahhh... inheritance :-)
=cut
   db::SendSQL("UPDATE dirent SET path = ". db::SqlQuote($fullyQualified) ." WHERE file_id = $file_id");
   return $file_id;
}

# Cache only the directory entries.
# Our cache will grow forever, but that will not be very big.
my %idCache;  # the key is "$parent-$fname" which stores a value of the child.

# This will determine file_id (used in the database) by filename.
# $createDirFlag will only create directories as needed.  $filename is assumed to be a directory.
sub id {
   my($filename,$createDirFlag) = @_;  
   $filename = "/" if $filename eq "";
   my $parent = 0;  # zero is the root directory
   my $oldParent = 0;
   for my $fname (split("/",$filename)) {
      next if $fname eq "";
      next if $fname eq ".";
      $oldParent = $parent;
      if($fname eq "..") {
         db::SendSQL("SELECT parent FROM dirent WHERE file_id = $parent");
         $parent = db::FetchOneColumn();
      } else {
         if(defined $idCache{"$parent-$fname"}) {  # if the query exists in the cache
             $parent = $idCache{"$parent-$fname"};
         } else {   # else the query is not already in the cache.
            db::SendSQL("SELECT file_id,type FROM dirent WHERE fname = ". db::SqlQuote($fname) . " AND parent = $parent");
            my($child,$type) = db::FetchSQLData();
            $idCache{"$parent-$fname"} = $child if(defined $type && $type eq "dir");
            $parent = $child;
         };
      };
      if(!defined($parent)) {
         if($createDirFlag) {
            $parent = createDir($fname,$oldParent);
         } else {
            $cgi::keyword{'warning'} = "<font size=+2 color=red>directory or file \"" .
                                  cgi::htmlQuote($fname) . "\" doesn't exist</font><br>\n";
            return -1;
         };
      };
   }
   return $parent;
}

sub cleanPath {
   my($path) = @_;

   $path =~ s|/+|/|g;  # eliminate repeated slashes
   while($path =~ s|/[^/]+/\.\./|/|g) {};  # eliminate parent dir references /../
   while($path =~ s|/\./|/|g) {};  # eliminate self dir references /./
   while($path =~ s|^\./||g) {};  # eliminate self dir references /./
   while($path =~ s|/\.$||g) {};  # eliminate self dir references /./
   #$path =~ s|./$||g;  # eliminate trailing slash
   return $path;
}

sub pathDepth {
# this function is broken
   my($path) = @_;
   return ($path =~ tr/\///) -1; # count slashes
}

sub admin_info {
   my($file_id) = @_;

   my %info;
   db::SendSQL("SELECT fname, parent, type, comment, head FROM " .
               "dirent WHERE file_id = " . db::SqlQuote($file_id)); 
   my @info = db::FetchSQLData();
   $info{'file_id'}  = $file_id;
   $info{'fname'}  = $info[0];
   $info{'parent'}  = $info[1];
   $info{'type'}  = $info[2];
   $info{'comment'}  = $info[3];
   $info{'head'}  = $info[4];
   return \%info;
}

#  $file_id is optional if the hash already has a file_id parameter (ala amin_info())
#
#  Okay, the return value is this complicated refs of refs crap.
#      $inforef{revision}{param} = value   
#  perl stores this as a hash containing hash references.
#
sub GetRevInfoFromDb {
   my($inforef,$file_id,$branch,$andwhere) = @_;

   $file_id = $inforef->{'file_id'} if !defined($file_id);
   my($branchspec1,$branchspec2) = ("","");
   if(defined($branch) and length($branch) and $branch ne "all") {
      $branchspec1 = "btime, ";
      $branchspec2 = "AND btime.file_id = revision.file_id AND branchname = " . db::SqlQuote($branch) . " AND ".
          "btime.branchnum = revision.branchnum AND (btime.timestamp IS NULL OR btime.timestamp >= revision.timestamp) ";
   };
   $andwhere ||= "";  # define if not defined
   db::SendSQL("SELECT revision, comment.author, state, branches, next, revision.timestamp, revision.branchnum, descr ".
               "FROM revision, $branchspec1 comment ".
               "WHERE revision.file_id = $file_id $branchspec2 AND comment.comment_id = revision.comment_id $andwhere"); 
   while(my @revision = db::FetchSQLData()) {
      my %revhash;
      $inforef->{$revision[0]} = \%revhash;
      $revhash{'author'}  = $revision[1];
      $revhash{'state'}  = $revision[2];
      $revhash{'branches'}  = $revision[3];
      $revhash{'next'}  = $revision[4];
      $revhash{'timestamp'}  = $revision[5];
      $revhash{'branchnum'}  = $revision[6];
      $revhash{'comment'}  = $revision[7];
   };
   return $inforef;
}

sub GetTagHashForFile {
   my($file_id) = @_;

   my(%tagHashRevKey,%tagHashNameKey);
   db::SendSQL("SELECT label, revision FROM label WHERE file_id = $file_id");
   while(my($label,$revision) = db::FetchSQLData()) {
      if(exists($tagHashRevKey{$revision})) {
         $tagHashRevKey{$revision} .= " $label";
      } else {
         $tagHashRevKey{$revision} = $label;
      };
      $tagHashNameKey{$label} = $revision;
   };
   return (\%tagHashRevKey,\%tagHashNameKey);
}

sub GetBranchHashForFile {
   my($file_id) = @_;

   my(%branchHashNumKey,%branchHashNameKey);
   db::SendSQL("SELECT branchname, branchnum FROM btime WHERE file_id = $file_id AND timestamp IS NULL");
   while(my($branchname,$branchnum) = db::FetchSQLData()) {
      #print("<h3>branchname=$branchname branchnum=$branchnum</h3>\n");
      if(exists($branchHashNumKey{$branchnum})) {
         $branchHashNumKey{$branchnum} .= " $branchname";
      } else {
         $branchHashNumKey{$branchnum} = $branchname;
      };
      $branchHashNameKey{$branchname} = $branchnum;
   };
   return (\%branchHashNumKey,\%branchHashNameKey);
}

sub GetBranchHeadHashForFile {
   my($file_id) = @_;

   my(%branchHeadHash);
   db::SendSQL("CREATE TEMPORARY TABLE latest2 ".
               "SELECT branchname, MAX(revision.rev_id) AS rev_id, acl FROM btime, revision ".
               "WHERE btime.file_id = revision.file_id AND btime.file_id = $file_id ".
               "AND btime.branchnum = revision.branchnum ".
               "AND (revision.timestamp <= btime.timestamp OR btime.timestamp IS NULL) ".
               "GROUP BY branchname");
   db::SendSQL("SELECT latest2.branchname, revision, acl FROM latest2, revision WHERE " . 
               "revision.file_id = $file_id AND revision.rev_id = latest2.rev_id ");
   while(my($branchname,$revision,$acl) = db::FetchSQLData()) {
      if(exists($branchHeadHash{$revision})) {
         $branchHeadHash{$revision} .= " $branchname"; 
      } else {
         $branchHeadHash{$revision} = $branchname; 
      }
   };
   db::SendSQL("DROP TABLE latest2");
   return \%branchHeadHash;
}

sub html_pwd {
   my($filename,$inforef) = @_; 
   $filename = cleanPath($filename); 
   my $reconstructed = "";
   print("<p>Current directory:\n");
   print(cgi::Link("/") . "root</a>\n");
   $filename =~ s/^\///;  # remove leading slash
   for my $dir (split('/',$filename)) {
      $reconstructed .= "/$dir";
      print("/" . cgi::Link("$reconstructed") . " $dir </a> ");
   }
}

sub GuessNextRev {
   my($cur) = @_;
   $cur =~ s/\.([^\.]+)$/"." . $1+1/e;  # increment last number
   return $cur;
}
sub GuessPreviousRev {
   my($cur) = @_;
   return "" if ! $cur;
   my @prevrev = split(/\./, $cur);
   if(--$prevrev[$#prevrev] <= 0) {
       # If it was X.Y.Z.1, just make it X.Y
       pop(@prevrev);
       pop(@prevrev);
   }
   return join(".", @prevrev);
}

# The $revref is a complicated hash of revision numbers, each one a hash of info, see GetRevInfoFromDb()
sub PreviousRev {
   my($cur,$revref) = @_;

   # Find the previous revision
   my @prevrev = split(/\./, $cur);
   my $prev;
   do {
       if (--$prevrev[$#prevrev] <= 0) {
           # If it was X.Y.Z.1, just make it X.Y
           pop(@prevrev);
           pop(@prevrev);
       }
       $prev = join(".", @prevrev);
   } until (defined($$revref{$prev}) || $prev eq "");
   return $prev;
}

sub rev_cmp($$)
{
   my($a,$b) = @_;
   my @a = split(/\./,$a);
   my @b = split(/\./,$b);
   while(1) {
      my $A = shift(@a);
      my $B = shift(@b);
      return 0 if(!defined($A) && !defined($B));
      return -1 if(!defined($A));
      return 1 if(!defined($B));
      my $cmp = ($A <=> $B);
      return $cmp if($cmp);
   }
}

sub Branches
{
   my($where) = @_;

   db::SendSQL("SELECT branchname FROM btime, dirent WHERE btime.file_id = dirent.file_id AND $where GROUP BY branchname");
   my @branches;
   while(my $branchname = db::FetchOneColumn()) {
      push(@branches,$branchname);
   }
   return @branches;
}

#  This function will return a list of listrefs.
#  $fields is a comma separated lists of fields from tables dirent, revision, or comment.
#
#  Okay, I tried MAX() over timestamp but we want to allow concurrent different branch
#  work and two branchs could have the same timestamp revision so the second select
#  needed a bunch more conditions to accomodate that.  Thus, we use rev_id, and it
#  allows multiple commits per second.
#
#  The $where clause written in SQL should only reference the dirent table, but the
#  revision table works for now.
#
#  "tags" are either branchname's or label's.  i.e. tag is the generalization of a branch or a label.
#  If you specify a label and it exists, this takes precident over branch stuff.
#  If the label doesn't exist, no revision is returned unless a branch is specified.
#  Except when a label matches, if you specify a branch which doesn't exist for each
#  file, then the file is not returned.
#
sub GetFiles 
{
   my($fields,$where,$order,$latest,$branch,$label) = @_;

   #print("<h2>latest $latest</h2>\n") if defined($latest);
   my $latespec = "";
   if(defined($latest) && $latest) {
      if($latest !~ /^\d+$/) { # if not simply an integer
         $latest = ::str2time($latest);
      };
      if(!defined($latest)) {
         $cgi::keyword{'warnings'} = "time \"" . cgi::htmlQuote($cgi::form{'latest'}). " not parsable<br>\n";
      } else {
         $latespec .= " AND revision.timestamp < $latest";
      };
   };
   my @rows;
   if(defined($branch) and length($branch) and $branch ne "all") {
      #print("<h2>branch $branch</h2>\n");
      $branch = db::SqlQuote($branch);
      # The bt2 table is because the acl is only valid when timestamp is null.
      db::SendSQL("CREATE TEMPORARY TABLE latest (primary key (file_id))".
                  "SELECT revision.file_id, MAX(revision.rev_id) AS rev_id, bt2.acl AS acl, bt2.owner AS owner ".
                  "FROM revision, dirent, btime ".  
                  "LEFT JOIN btime AS bt2 ON bt2.file_id = btime.file_id AND bt2.branchname = btime.branchname ".
                  "AND bt2.branchnum = btime.branchnum AND bt2.timestamp IS NULL ".
                  "WHERE revision.file_id = dirent.file_id AND $where $latespec ".
                  "AND btime.file_id = dirent.file_id AND btime.branchname = $branch ".
                  "AND (revision.timestamp <= btime.timestamp OR btime.timestamp IS NULL) AND revision.branchnum = btime.branchnum ".
                  "GROUP BY revision.file_id");
   } elsif(defined $label && length $label) {  # else only from the one label
      # steal the table definition but insert no rows.
      db::SendSQL("CREATE TEMPORARY TABLE latest (primary key (file_id))".
                  "SELECT revision.file_id, revision.rev_id AS rev_id, btime.acl, btime.owner ".
                  "FROM revision, dirent ".  
                  "LEFT JOIN btime ON revision.file_id = btime.file_id AND revision.branchnum = btime.branchnum ".
                  "WHERE 0");
   } else {  # else any branch
      db::SendSQL("CREATE TEMPORARY TABLE latest (primary key (file_id))".
                  "SELECT revision.file_id, MAX(revision.rev_id) AS rev_id, btime.acl, btime.owner ".
                  "FROM revision, dirent ".  
                  "LEFT JOIN btime ON revision.file_id = btime.file_id AND revision.branchnum = btime.branchnum ".
                  "WHERE revision.file_id = dirent.file_id AND $where $latespec ".
                  "AND dirent.file_id = btime.file_id AND btime.timestamp IS NULL ".
                  "GROUP BY revision.file_id ");
   };
   # update for labels
   db::SendSQL("REPLACE latest SELECT label.file_id, revision.rev_id, btime.acl, btime.owner ".
               "FROM label, dirent, revision ".
               "LEFT JOIN btime ON label.file_id = btime.file_id AND revision.branchnum = btime.branchnum AND btime.timestamp IS NULL ".
               "WHERE label.file_id = dirent.file_id AND label.file_id = revision.file_id ".
               "AND label.revision = revision.revision AND label.label = ". db::SqlQuote($label) ."AND $where");
   # include all the dirent's which are not version controlled (have no revisions)
   db::SendSQL("INSERT latest SELECT dirent.file_id, NULL AS rev_id, dirent.acl, NULL AS owner ".
               "FROM dirent LEFT JOIN revision ".
               "USING (file_id) WHERE $where AND revision.file_id IS NULL");
   db::SendSQL("SELECT $fields FROM dirent ".
                 "INNER JOIN latest USING (file_id) ".
                 "LEFT JOIN revision USING (rev_id, file_id) ".
                 "LEFT JOIN comment USING (comment_id) ". 
                 "WHERE $where ORDER BY $order");
   @rows = db::FetchListOfLists();
   db::SendSQL("DROP TABLE latest");
   return @rows;
}

sub applyDiffs
{
   my($diff,$textLinesRef) = @_;
   my @diffLines = split(/^/m,$diff);
   my $grown = 0;
   while(my $command = shift(@diffLines)) {
      if($command =~ /^a(\d+)\s+(\d+)/) {
         my($where,$howMany) = ($1,$2); 
         splice(@$textLinesRef,$where+$grown,0,splice(@diffLines,0,$howMany,()));
         $grown += $howMany;
      }
      if($command =~ /^d(\d+)\s+(\d+)/) {
         my($where,$howMany) = ($1,$2); 
         splice(@$textLinesRef,$where+$grown-1,$howMany,());
         $grown -= $howMany;
      }
   }
}

# Each line is a list element (containing newline)
sub GetText
{
   my($file_id,$revision) = @_;
 
   db::SendSQL("SELECT rev_id FROM revision WHERE ".
               "revision = ".db::SqlQuote($revision)." AND file_id = $file_id");
   my $rev_id = db::FetchOneColumn();
   if(!defined($rev_id)) {
      return undef;
   };
   # first get the tip of the tree version
   db::SendSQL("SELECT MAX(rev_id) FROM revision WHERE file_id = $file_id AND branchnum = ''");
   my $head_rev_id = db::FetchOneColumn();
   db::SendSQL("SELECT text, revision FROM revision WHERE rev_id = $head_rev_id");
   my($text,$revnum) = db::FetchSQLData();
   if($revnum eq $revision) {
      return split(/^/m,$text) if wantarray;
      return $text;
   };
   my @textLines = split(/^/m,$text);
   # then work our way down the MAIN branch
   my @nums = split(/\./,$revision);
   $revision =~ m/^([^\.]+\.[^\.]+)/;  # first two numbers of revision
   my $trunkBranchRevNum = $1;
   db::SendSQL("SELECT rev_id FROM revision WHERE ".
               "revision = ".db::SqlQuote("$nums[0].$nums[1]")." AND file_id = $file_id");
   my $trunkBranchRevId = db::FetchOneColumn();
   db::SendSQL("SELECT text,rev_id,revision FROM revision WHERE rev_id >= $trunkBranchRevId AND ".
               "file_id = $file_id AND branchnum = '' AND rev_id < $head_rev_id ORDER BY -rev_id");
   while(my($textdiff,$rev_id,$rr) = db::FetchSQLData()) {
      applyDiffs($textdiff,\@textLines);
   };
   #return join("\n",@textLines) if($trunkBranchRevNum eq $revision);
   # then walk each branch
   my $numNumsUsed = 2;
   while(1) {  # for each branch to walk
      last if($numNumsUsed > $#nums);
      $numNumsUsed += 2;
      my $branchNum = join(".",@nums[0..$numNumsUsed-2]);
      my $branchEnd = join(".",@nums[0..$numNumsUsed-1]);
      db::SendSQL("SELECT rev_id FROM revision WHERE ".
                  "revision = ".db::SqlQuote($branchEnd)." AND file_id = $file_id");
      my $branchEndId = db::FetchOneColumn();
      db::SendSQL("SELECT text,rev_id,revision FROM revision WHERE rev_id <= $branchEndId AND ".
                  "file_id = $file_id AND branchnum = ". db::SqlQuote($branchNum) ." ORDER BY rev_id");
      while(my($textdiff,$rev_id,$rr) = db::FetchSQLData()) {
         applyDiffs($textdiff,\@textLines);
      };
   }
   return @textLines if wantarray;
   return join("",@textLines);
}

1;















