#!/usr/bin/perl -w

# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is the Bugzilla Bug Tracking System.
#
# The Initial Developer of the Original Code is Netscape Communications
# Corporation. Portions created by Netscape are
# Copyright (C) 1998 Netscape Communications Corporation. All
# Rights Reserved.
#
# Contributor(s): Terry Weissman <terry@mozilla.org>
#                 Dan Mosedale <dmose@mozilla.org>
#                 Jake <jake@acutex.net>

# Contains some global variables and routines used throughout bugzilla.

#use strict;
#use diagnostics;

package db;

use DBI;
use Carp;
use config;

sub die_with_dignity {
    my ($err_msg) = @_;
    print $err_msg;
    confess($err_msg);
}
$::SIG{__DIE__} = \&die_with_dignity if($config::debug);

my $db;

sub ConnectToDatabase {
    return if defined($db);

    $db = DBI->connect("DBI:mysql:host=$config::db_host;database=$config::db_name", $config::db_user, $config::db_pass)
        || die "cvsdb is currently broken. Please try again later. " . 
            "If the problem persists, please contact " .
            ". The error you should quote is: " . $DBI::errstr;
}

# This is used to manipulate global state used by SendSQL(),
# MoreSQLData() and FetchSQLData().  It provides a way to do another
# SQL query without losing any as-yet-unfetched data from an existing
# query.  Just push the current global state, do your new query and fetch
# any data you need from it, then pop the current global state.
# 
my @SQLStateStack = ();

my @fetchahead;
my $currentquery;


sub PushGlobalSQLState() {
    push @SQLStateStack, $currentquery;
    push @SQLStateStack, [ @fetchahead ]; 
}

sub PopGlobalSQLState() {
    die ("PopGlobalSQLState: stack underflow") if ( $#SQLStateStack < 1 );
    @fetchahead = @{pop @SQLStateStack};
    $currentquery = pop @SQLStateStack;
}

sub SavedSQLStates() {
    return ($#SQLStateStack + 1) / 2;
}

sub SendSQL {
    my ($str) = (@_);
    if($config::debug && -w "/tmp/dcvs-sql-log") {
       my $prettysql = $str;
       $prettysql =~ s/(WHERE|AND|LEFT JOIN|INNER JOIN|JOIN|FROM|GROUP|ORDER)/\n$1/g;
       open(SQLLOG,">>/tmp/dcvs-sql-log");
       print(SQLLOG scalar(localtime()), "\n");
       print(SQLLOG $prettysql, "\n\n");
       close(SQLLOG);
    };
    ConnectToDatabase() unless $db;
#print STDERR "###&&& $str\n";
    $currentquery = $db->prepare($str);
    $currentquery->execute || die "$str: " . $db->errstr;
    my $q = SqlQuote($str);
}

sub MoreSQLData {
    if (@fetchahead) {
	return 1;
    }
    if (@fetchahead = $currentquery->fetchrow_array) {
	return 1;
    }
    return 0;
}

sub FetchSQLData {
    if (@fetchahead) {
	my @result = @fetchahead;
	undef @fetchahead;
	return @result;
    }
    return $currentquery->fetchrow_array;
}

sub FetchListOfLists {
   my @lol = (); 
   while(my @list = FetchSQLData) {
      push(@lol,\@list);
   }
   return @lol;
}

sub FetchOneColumn {
    my @row = FetchSQLData();
    return $row[0];
}

sub last_insert_id
{
   SendSQL("SELECT LAST_INSERT_ID()");
   return FetchOneColumn();
}


# This routine is largely copied from Mysql.pm.

sub SqlQuote {
    my ($str) = (@_);
#     if (!defined $str) {
#         confess("Undefined passed to SqlQuote");
#     }
    return "''" if !defined($str);
    $str =~ s/([\\\'])/\\$1/g;
    $str =~ s/\0/\\0/g;
    # If it's been SqlQuote()ed, then it's safe, so we tell -T that.
    #$str = detaint_string($str);
    return "'$str'";
}

our $now;

sub getDbNowTime
{
   my($printWarningFlag) = @_;

   db::SendSQL("SELECT UNIX_TIMESTAMP(NOW())");
   $now = db::FetchOneColumn();
   if($printWarningFlag) {
      my $perltime = time();
      if($perltime > $now + 2 || $perltime < $now - 2) {
         print("<br><font color=red>WARNING: database and webserver clocks differ by ".($perltime-$now)." seconds. \n".
               "I will use database time.</font>");
      }
   };
   return $now;
}

1;















