#!/usr/bin/perl -w -T
use strict;
#use diagnostics;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>


# Add the directory of the executable to the module search path.
BEGIN { unshift @INC, "/var/www/html/dcvs" };

use config;
use db;
use cvsdb;
use session;
use cdiff;
use unixdiff;
use rcsdiff;
use merge;
use Date::Format;          
use Date::Parse;        
use IO::Socket;
use Getopt::Long;
use Digest::MD5 qw(md5_base64);

$ENV{'PATH'} = '/bin:/usr/bin';
delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};

my $nofork;
my $daemon;
my $localaddr = undef;   
my $port = 2401;   
GetOptions("nofork" => \$nofork,
           "daemon" => \$daemon,
           "localaddr=s" => \$localaddr,
           "port=i" => \$port) or 
   die("usage dcvsd.pl [options]   
                 --nofork 
                 --daemon 
                 --localaddr=0.0.0.0 
                 --port=2401\n");
die("You must have either --nofork or --daemon\n") if !defined $nofork && !defined $daemon;
die("You cannot have both --nofork and --daemon\n") if defined $nofork && defined $daemon;

# Here is the big parsing state.  
# Please study the protocol first (cvsclient.ps)
my($username,$validResponses);
my @args;  # Collects the "Arguement" statements.  Order matters.
my $lastEntry;  # This came from the last "Entries" line.
                # It uses the repository path.
my $root;  # This has the $config::root prefix removed
my $localdir;   # This is from the last "Directory" received.
            # It is the client-side "local directory".
            # It will not begin or end with a slash.
my $rdir;  # This is from the last "Directory" received. 
           # It has the $config::root prefix removed.
           # It will begin with a slash but not end with a slash.
# The following are from flags for checkout or update commands
my($createDirsFlag,$clearStickyFlag);
# The following is the database sent to us about directories.
my(%lDir,%stickyDir,%staticDir); # The index is $rdir
# The following is the database sent to us about files
my(%rev,%kopt,%checkinTime,%unchanged,%mode,@questionable,%conflict,%sticky); # The index is "$rdir/$fname"
my(%text); # This is from "Modified" statements
# The following is a hash of hashes $toadd{$repos}{$fname}
my(%toadd);

sub resetStateForNofork {  # if we didn't fork
   ($root,$username,$validResponses,@args,$localdir,$rdir,$lastEntry) = undef;
   ($createDirsFlag,$clearStickyFlag) = undef;
   (%lDir,%stickyDir,%staticDir) = (); 
   (%rev,%kopt,%checkinTime,%unchanged,%mode,%text,@questionable,%conflict,%sticky) = (); 
   (%toadd) = ();
}

my($sock,$peer);

sub sockprint
{
   if($nofork) {
      my($text) = @_;
      $text =~ s/^/---/gm;   # append prefix
      print("$text");
   };
   $sock->print($_[0]);
}

sub readsock
{
   my $line = $sock->getline() || "";
   print(">>>$line") if $nofork;
   chomp($line);
   return $line;
}

sub bitchslap 
{
   sockprint("E dcvs server: fatal! " . $_[0] . "\n");
   close $sock;
}

# The cypher is a single symmetric substitution.  
# The encrypt and decrypt are the same.
sub pserverCrypt
{
   my($x) = @_;

   my @secretDecoderRing = 
         (114,120,53,79,96,109,72,108,70,64,76,67,116,74,68,87,111,52,75,119,
          49,34,82,81,95,65,112,86,118,110,122,105,41,57,83,43,46,102,
          40,89,38,103,45,50,42,123,91,35,125,55,54,66,124,126,59,47,92,71,
          115,78,88,107,106,56,36,121,117,104,101,100,69,73,99,63,94,93,39,
          37,61,48,58,113,32,90,44,98,60,51,33,97,62,77,84,80,85);
   $x =~ s/([ -~])/pack("C",$secretDecoderRing[unpack("C",$1)-32])/eg;
   return $x;
}

sub doAuthenticate 
{
   print("doClientSession()\n") if $nofork;
   my($begin,$password,$end);
   if(!defined($begin = $sock->getline()) ||
      !defined($root = $sock->getline()) ||
      !defined($username = $sock->getline()) ||
      !defined($password = $sock->getline()) ||
      !defined($end = $sock->getline())) {
      print("could not read authentication\n");
   };
   if($begin !~ /^BEGIN (AUTH|VERIFICATION) REQUEST/ || 
      $end !~ /^END (AUTH|VERIFICATION) REQUEST/) {
      print("bad syntax during authorization\n");
      sockprint("E bad syntax during authorization\n");
   };
   chomp($password);
   chomp($username);
   chomp($root);
   if($password =~ s/^A//) {
      $password = pserverCrypt($password);
   }  # else no encoding?
   my($userid,undef,undef,undef,$group) = session::isPasswordValid($username,$password);
   if(! defined $userid) {
      sockprint("I HATE YOU\n");  # bad password
      return 0;  # return "failure"
   }
   $session::user_id = $userid;   # save in global for session::isPrivilegedTo()
   session::updateGroupGlobal($group);
   sockprint("I LOVE YOU\n");  # good password
   substr($root,0,length($config::root),"") eq $config::root or  # strip prefix and check it
      bitchslap("bad root (should be at least $config::root)");
   return 1;  # return "success"
}

sub doRoot {
   $root = $_[0];
   substr($root,0,length($config::root),"") eq $config::root or
      bitchslap("bad root");
}
sub doValidResponses {
   $validResponses = $_[0];
}
sub doValidRequests {
   sockprint("Valid-requests Root Valid-responses valid-requests Directory Max-dotdot Static-directory ".
                "Sticky Checkin-prog Update-prog Entry Kopt Checkin-time Checkin-prog Update-prog Entry ".
                "Kopt Checkin-time Modified Is-modified Unchanged UseUnchanged Notify Questionable ".
                "Case Argument Argumentx Global_option Set ci diff tag status ".
#"expand-module ".
                "admin history watchers editors annotate log co export rdiff rtag init update ".
                "import add remove watch-on watch-off watch-add watch-remove release noop ".
                "noop update-patches wrapper-scheme-rcsOptions version\nok\n");
   @args = ();
}
sub doVersion {
   sockprint("M dcvs v0.0\nok\n");
}
sub doUseUnchanged {
}
sub doNoop {
   sockprint("ok\n");
}
sub doDirectory {
   $localdir = $_[0];
   $rdir = readsock();
   # prune the required root
   substr($rdir,0,length($config::root),"") eq $config::root or
      bitchslap("bad root prefix");
   $lDir{$rdir} = $localdir;
}
sub doSticky {
   $stickyDir{$rdir} = $_[0];
}
# Static directores do not add files.  
# For example in "cvs co foo/baz/bam/"  foo/ and foo/baz/ are static
sub doStaticDirectory {
   $staticDir{$rdir} = $_[0];
}
sub doEntry {
   my(undef,$filename,$revision,$conflict,$options,$sticky,$junk) = split("/",$_[0],7);
   $lastEntry = "$rdir/$filename";  # save in global variable
   $rev{$lastEntry} = $revision;
   $conflict{$lastEntry} = $conflict if length($conflict);
   $kopt{$lastEntry} = $options if length($options);
   $sticky{$lastEntry} = $sticky if length($sticky);
   $toadd{$rdir}{$filename} = 1 if $revision eq "0";
}
sub doUnchanged {
   $unchanged{"$rdir/$_[0]"} = 1;
}
sub doKopt {
   $kopt{$lastEntry} = $_[0];
}
# We shall ignore check in time.
sub doCheckinTime {
   # We intentionally ignore this!
   #$checkinTime{"$rdir/$fname"} = $_[0];
}
sub doModified {
   my $fname = $_[0];

   $mode{"$rdir/$fname"} = readsock();
   my $size = readsock();
   if($size > $config::maxFileSize) {
      sockprint("E dcvs server: $rdir/$fname file too large ($size > $config::maxFileSize).\nerror\n");
      $sock->close;
   } else {
      $sock->read(my $text,$size) ;
      $text{"$rdir/$fname"} = $text;
   }
}
sub doIsModified {
}
sub doQuestionable {
   push(@questionable,$lastEntry);
}

sub doArgument {
   push(@args,$_[0]);
}
# Argumentx is a continuation of the previous argument with a newline.
# For example sending a commit comment which contains newlines.
sub doArgumentx {
   $args[-1] .= "\n$_[0]";  # append to previous arg, last in list
}

# This is called for every file the database says they should/will have with an update or checkout.
sub ProcessSingleFile {
   my($file_id,$fname,$repoRev,$ldir,$repos,$command,$sticky,$state,$repoMode,$timestamp) = @_;

   #my $keyname = cvsdb::cleanPath("$repos/$fname");
   my $keyname = "$repos/$fname";
#printf("#### $keyname  $repoRev %s\n",$rev{$keyname} || "");
   delete $toadd{$repos}{$fname};  # doesn't need to be added if already exists in the repository
   if($command eq "co" || (! exists $unchanged{$keyname} && ! exists $text{$keyname})) { 
      if($state ne "dead") {
         my $targetText = cvsdb::GetText($file_id,$repoRev);
         my $length = length($targetText);
         sockprint("Mod-time ". time2str("%e %b %Y %H:%M:%S -0000",$timestamp,"UTC") . "\n" .
                   "MT text U \nMT fname $fname\nMT newline\n".
                   "Created $ldir/\n$keyname\n/$fname/$repoRev///$sticky\n$repoMode\n$length\n");
         $sock->print($targetText);
      }
   } elsif($repoRev eq $rev{$keyname} && defined $unchanged{$keyname}) { # else version is correct and unchanged
      if(($sticky{$keyname}||"") ne $sticky) {  # else not modified, version correct, but if wrong sticky 
         sockprint("Checked-in $ldir/\n$keyname\n/$fname/$repoRev///$sticky\n");
      }  
   } else { # else command is an update and a file exists
      my $oldtext = cvsdb::GetText($file_id,$rev{$keyname}) if exists $text{$keyname};
      if(exists $text{$keyname} && !defined($oldtext)) {  # if it is modified but we can't find from what
         sockprint("E dcvs server: revision $rev{$keyname} of $keyname is unknown\n");
      } elsif(exists $text{$keyname} && $text{$keyname} ne $oldtext) {  # if was presented as "Modified" and really is
          if($state eq "dead") {  
             sockprint("E dcvs server: conflict: $fname is modified but no longer in the repository\n".
                       "MT text C\nMT fname $fname\nMT newline");
          } elsif($rev{$keyname} eq $repoRev) {  # if exists modified from correct version
             sockprint("MT text M \nMT fname $fname\nMT newline\n");
          } else {  
             sockprint(#"M RCS file: $keyname\nM retrieving revision $rev{$keyname}\nM retrieving revision $repoRev\n".
                       #"M Merging differences between $rev{$keyname} and $repoRev into $fname\n".
                       "Copy-file $ldir/\n$keyname\n.#$fname.$rev{$keyname}\n".
                       "M   Your modified version saved as .#$fname.$rev{$keyname}\n");
             my $targetText = cvsdb::GetText($file_id,$repoRev);
             my($mergetext,$reject) = merge::merge($oldtext,$targetText,$text{$keyname}); # port the patch
             my $length = length($mergetext);
             if(length $reject) {
                $reject = "--- $keyname ". ::time2str("%C",$timestamp,"UTC") . "    $rev{$keyname}\n" .
                          "+++ $keyname ". ::time2str("%C",$timestamp,"UTC") . "\n" . $reject;
                sockprint("Merged $ldir/\n$keyname\n/$fname/$rev{$keyname}///\n$mode{$keyname}\n". length($reject) ."\n");
                $sock->print($reject);   # so it doesn't show in the debug output
                sockprint("Copy-file $ldir/\n$keyname\n$fname.$rev{$keyname}.rej\n".
                          "M   Rejected hunks saved as $fname.$rev{$keyname}.rej\n");
                sockprint("Merged $ldir/\n$keyname\n/$fname/$repoRev/+=//$sticky\n$mode{$keyname}\n$length\n");
                $sock->print($mergetext);   # so it doesn't show in the debug output
                sockprint("MT text C \nMT fname $fname\nMT newline\n");
             } else {
                sockprint("Merged $ldir/\n$keyname\n/$fname/$repoRev///$sticky\n$mode{$keyname}\n$length\n");
                $sock->print($mergetext);   # so it doesn't show in the debug output
                sockprint("MT text M \nMT fname $fname\nMT newline\n");
             };
          };
      } elsif($state eq "dead") {  # else not modified, and is going away.
         sockprint("E dcvs server: $fname is no longer in the repository ($repoRev is dead)\n".
                   "Removed $ldir/\n$keyname\n");
      } elsif($repoRev ne $rev{$keyname}) {  # else not modified, but if version is wrong
         my $targetText = cvsdb::GetText($file_id,$repoRev);
         my $length = length($targetText);
         sockprint("Mod-time ". time2str("%e %b %Y %H:%M:%S -0000",$timestamp,"UTC") . "\n" .
                   "MT text U \nMT fname $fname\nMT newline\nUpdate-existing $ldir/\n$keyname\n".
                   "/$fname/$repoRev///$sticky\n$repoMode\n$length\n");
         $sock->print($targetText);
      } elsif(($sticky{$keyname}||"") ne $sticky) {  # else not modified, version correct, but if wrong sticky 
         sockprint("Checked-in $ldir/\n$keyname\n/$fname/$repoRev///$sticky\n");
      };
   }
}

sub doCheckoutOrUpdate {
   my(undef,$command) = @_;
   my($latest,$branch,$label,$sticky,$reposNickname,$noCollapse) = (undef,"MAIN",undef,"","");
   while(my $arg = shift(@args)) {
      if($arg eq "-N") {
         $noCollapse++;
      } elsif($arg eq "-A") {  # clear sticky tags
         $clearStickyFlag++;
      } elsif($arg eq "-P") {  # seen on checkout.  meaning?
      } elsif($arg eq "-u") {  # seen no update.  meaning?
      } elsif($arg eq "-d") {  # for update, create missing directories
         if($command eq "co") {
            $reposNickname = shift(@args);
         } else {  # else is an update
            $createDirsFlag++;
         };
      } elsif($arg eq "-D") {  # date modifier, too bad it doesn't pass the time zone
         $latest = str2time(shift(@args)) or bitchslap("bad date format");
         $sticky = $sticky || "D" . time2str("%Y.%m.%d.%H.%M.%S",$latest,"UTC");  # branch preferred
      } elsif($arg eq "-r") {  # tag (branch or label) specifier 
         $label = $branch = shift(@args);
         $sticky = "T" . $branch;
      } else {
         unshift(@args,$arg);
         last;
      }
   }
   sockprint("E dcvs server: -N (directory nickname non-collapse) not supported\n") if $noCollapse && $reposNickname;
   $command eq "co" and $createDirsFlag++;
   my @triples = ();
   my %createdNonRecurseDirs;
   push(@args,".") if $command eq "update" and $#args == -1;  # do local dir if nothing specified for update 
   for my $arg (@args) {  # for each item we are asked to check out
      $arg = cvsdb::cleanPath($arg);
      my $id = cvsdb::id("$rdir/$arg");
      if($id == -1) {
         sockprint("E dcvs server: $rdir/$arg is not in the database.\n"); 
         next;
      };
      if($arg eq ".") {
         push(@triples,[ "$localdir", "$rdir", "dirent.parent = $id", 1 ]);
         next;
      } 
      $arg = cvsdb::cleanPath($arg);
      if($reposNickname) {  # if localdir has a slash
         push(@triples,[ ".", "$rdir", "dirent.file_id = $id", 1 ]);
         next;
      } 
      my $argPrefix = "";
      $argPrefix = "/$1" if $arg =~ m|(.+)/([^/]+)/?$|;  # match path and last file/dir
      push(@triples,[ "$localdir$argPrefix", "$rdir$argPrefix", "dirent.file_id = $id", 1 ]);
      while(length $argPrefix && $command eq "co") {  # for every leading directory we need to create (which doesn't recusively checkout)
         last if $createdNonRecurseDirs{$argPrefix}++ > 1;
         $arg = $argPrefix;
         $argPrefix = "";
         $argPrefix = $1 if $arg =~ m|(.+)/([^/])+/?$|;
         # We need to mention these directores so the client creates them and a CVS/ 
         unshift(@triples,[ "$localdir$argPrefix", "$rdir$argPrefix", "dirent.file_id = " . cvsdb::id("$rdir/$arg"), 0 ]);
      };
   }
   while(my $triple = shift(@triples)) {  # for each directory we will populate
      my($ldir,$repos,$where,$recurse) = @$triple;
      #print("TRIPLE {$ldir} {$repos} {$where} {$recurse}\n") if $config::debug;
      $ldir = cvsdb::cleanPath($ldir);
      $repos = cvsdb::cleanPath($repos);
      my @files = cvsdb::GetFiles("fname,revision.revision,type,dirent.file_id,state,dirent.mode,".
                                  "revision.timestamp,latest.acl,latest.owner",
                                  $where,"fname",$latest,$branch,$label);
      sockprint("E dcvs server: Updating $ldir/\n") if $where =~ /^parent/;
      while(my $fileref = shift(@files)) {  # each dirent of last query
         my($fname,$repoRev,$type,$file_id,$state,$repoMode,$timestamp,$acl,$owner) = @$fileref;
         next unless session::isPrivilegedTo($acl,0x0004,$owner);  # skip if not visible visible
         if(! session::isPrivilegedTo($acl,0x0008,$owner)) {  # if not read-f "file"
            sockprint("E Read privileges failed on file $fname tag $branch\n");
            next;
         };
         if($type eq "dir") {
            my $nickname = $fname;
            $nickname = $reposNickname if length($reposNickname) && $ldir eq ".";
            if($sticky) {
               sockprint("Set-sticky $ldir/$nickname/\n$config::root$repos/$fname/\n$sticky\n");
            } elsif($stickyDir{$repos} && !$clearStickyFlag) { 
               sockprint("Set-sticky $ldir/$nickname/\n$config::root$repos/$fname/\n$stickyDir{$repos}\n");
            } else { 
               sockprint("Clear-sticky $ldir/$nickname/\n$config::root$repos/$fname/\n");
            }
            if(!$recurse || ($staticDir{"$repos/$fname"} && !$createDirsFlag)) {
               sockprint("Set-static-directory $ldir/$nickname/\n$config::root$repos/$fname/\n");
            } else {
               sockprint("Clear-static-directory $ldir/$nickname/\n$config::root$repos/$fname/\n");
               push(@triples,[ "$ldir/$nickname", "$repos/$fname", "parent = $file_id", 1 ]) if $recurse;
            }
         } elsif($type eq "file") {
            ProcessSingleFile($file_id,$fname,$repoRev,$ldir,$repos,$command,$sticky,$state,$repoMode,$timestamp);
         };
      };
      # We need to check for new adds.  If the file were already in the repository it would be removed from %toadd
      if(defined $toadd{$repos}) {
         foreach my $fname (keys %{$toadd{$repos}}) {
            sockprint("MT text A \nMT fname $fname\nMT newline\n");
         };
      };
   };
   sockprint("ok\n");
}

# Why does it send us "Entry" info?  We ignore it.
sub doLog {
   my($headonly,$omitnames,$rcsonly);
   while(my $flag = shift @args) {
      if($flag eq "-h") {
         $headonly++; 
      } elsif($flag eq "-N") {
         $omitnames++; 
      } elsif($flag eq "-R") {
         $rcsonly++; 
      } else {
         unshift(@args,$flag);
         last;
      } 
   }   
   my @args = map { defined $rev{"$rdir/$_"} ? () : "$rdir/$_" } @args;  # keep args if they don't have $rev{}
   foreach my $keyname (sort keys(%rev),@args) {  # for each Entry line the client sent
      my $id = cvsdb::id($keyname);
      if($id == -1) {
         sockprint("E dcvs server: nothing known about $config::root$keyname\n");
         next;
      };
      if($rcsonly) {
         sockprint("M $config::root$keyname,v\n");
         next;
      };
      db::SendSQL("SELECT head FROM dirent WHERE file_id = $id");
      my $head = db::FetchOneColumn();
      $keyname =~ /\/?([^\/]*)$/;  # keep filename after last slash.
      sockprint("M \n".
                "M RCS file: $config::root$keyname,v\n".
                "M Working file: $1\n".
                "M head: $head\n");
      if(! $omitnames) {
         sockprint("M symbolic names:\n");
         db::SendSQL("SELECT label,revision FROM label WHERE file_id = $id"); 
         while(my($label,$repoRev) = db::FetchSQLData()) {
            sockprint("M        $label: $repoRev\n"); 
         }
         db::SendSQL("SELECT branchname,branchnum FROM btime WHERE file_id = $id AND timestamp IS NULL"); 
         while(my($branchname,$repoRev) = db::FetchSQLData()) {
            next if $branchname eq "MAIN";  
            sockprint("M        $branchname: $repoRev\n"); 
         }
      };
      if(! $headonly) {
         db::SendSQL("SELECT COUNT(revision) FROM revision WHERE file_id = $id");  # count the number of revisions
         my $totRevCount = db::FetchOneColumn();
         db::SendSQL("SELECT revision, revision.timestamp, state, comment.descr, comment.author, added, removed ".
                     "FROM revision LEFT JOIN comment USING (comment_id) ".
                     "WHERE file_id = $id AND branchnum = '' ".
                     "ORDER BY - rev_id");  # get trunk
         my @rows =  db::FetchListOfLists();
         db::SendSQL("SELECT revision, revision.timestamp, state, comment.descr, comment.author, added, removed ".
                     "FROM revision LEFT JOIN comment USING (comment_id) ".
                     "WHERE file_id = $id AND branchnum != '' ".
                     "ORDER BY branchnum, - rev_id");  # get branches
         push(@rows,db::FetchListOfLists());
         my $numRevSelected = $#rows+1;
         sockprint("M totoal revisions: $totRevCount;  selected revisions: $numRevSelected\n".
                   "M descriptions:\n");
         while(my $rowref = shift @rows) {
            my($repoRev,$ts,$state,$descr,$author,$numAdded,$numRemoved) = @$rowref;
            $descr =~ s/^/M /gm;  # prefix every line with "M "
            $descr .= "\n" if $descr !~ /\n$/;  # append trailing newline if not already present.
            my $date = ::time2str("%Y/%m/%d %H:%M:%S UTC",$ts,"UTC");
            ($numAdded,$numRemoved) = qw(? ?) if ! defined $numAdded;
            sockprint("M ----------------------------\n".
                      "M revision $repoRev\n".
                      "M date: $date;   author: $author;    state: $state;  lines: +$numAdded -$numRemoved\n".
                      $descr);
         };
      };
      sockprint("M =============================================================================\n");  
   };
   sockprint("ok\n");
}

sub doRemove {
   foreach my $arg (@args) {
      $arg =~ /(.*?)\/?([^\/]+)$/;  # split into path and filename 
      my($path,$filename) = ($1,$2);
      my $keyname = "$rdir/$arg";
      if(! defined $rev{"$keyname"}) {
          sockprint("E dcvs server: no Entry line for $arg\n");
      } else {
         $path = "." if $path eq "";
         my $sticky = $sticky{$keyname} || "";
         sockprint("E dcvs server: scheduling file '$arg' for removal\n".
                   "Checked-in $path/\n".
                   "$config::root$rdir/$arg\n".
                   "/$filename/-$rev{$keyname}///$sticky\n");
      };
   };
   sockprint("E dcvs server: use 'cvs commit' to remove this file permanently\n".
             "ok\n");
}

# This is files only (no directores) right now.
sub doAdd {
   foreach my $arg (@args) {
      $arg =~ /(.*?)\/?([^\/]+)$/;  # split into path and filename 
      my($path,$filename) = ($1,$2);
      my $keyname = "$rdir/$arg";
      if($lDir{"$rdir/$arg"}) {
         my $parent_id = cvsdb::id("$rdir/$path");
         if($parent_id == -1) {
            sockprint("E dcvs server: could not find parent dir $config::root$rdir/$path for $filename\n");
         } else {
            my $dir_id = cvsdb::createDir($filename,$parent_id);
            sockprint("M dcvs server: Directory $config::root$rdir/$arg added to repository\n");
            sockprint("M dcvs server: --> Using per-directory sticky tag '$stickyDir{$rdir . $arg}'\n") if $stickyDir{"$rdir/$arg"};
         }
      } else {
         $path = "." if $path eq "";
         my $sticky = $sticky{$keyname} || "";
         sockprint("E dcvs server: scheduling file '$arg' for addition\n".
                   "Checked-in $path/\n".
                   "$config::root$rdir/$arg\n".
                   "/$filename/0///$sticky\n");
      };
   };
   sockprint("E dcvs server: use 'cvs commit' to add this file permanently\n".
             "ok\n");
}

sub doDiff {
   my($context) = ("");
   while(my $flag = shift @args) {
      if($flag eq "-u") {
         $context = "-u";
      } else {
         unshift(@args,$flag);
         last;
      } 
   }   
   foreach my $keyname (sort keys(%rev)) { # for each Entry line the client sent
      next if(! defined $text{$keyname});
      if($rev{$keyname} eq "0") {
         (my $fname = $keyname) =~ s/.*\///s;  # delete last slash and before
         sockprint("E dcvs server: $fname is a new entry, no comparison available\n");
         next;
      }
      my $id = cvsdb::id($keyname);
      my @text1 = split(/^/m,$text{$keyname});
      my @text2 = cvsdb::GetText($id,$rev{$keyname});
      if(defined $text2[0]) {
         $keyname =~ /([^\/]+)$/;
         sockprint("M ===================================================================\n".
                   "M RCS file: $config::root$keyname,v\n".
                   "M retrieving revision $rev{$keyname}\n".
                   "M diff $context -r$rev{$keyname} $keyname\n");
         db::SendSQL("SELECT timestamp FROM revision WHERE file_id = $id AND revision = " . db::SqlQuote($rev{$keyname}));
         my $mtime = db::FetchOneColumn() || 0;
         my $diff;
         if($context) {
            $diff = cdiff::cdiff(\@text2,\@text1,$keyname,$keyname,$mtime,db::getDbNowTime(),"M ");
         } else {
            $diff = unixdiff::unixdiff(\@text2,\@text1,"M ");
         }
         sockprint($diff);
      } else {
         sockprint("E unknown revision $rev{$keyname} of $keyname\n");
      };
   }
   sockprint("ok\n");
}

# The return value is a list, where the end contains your requested fields.
# Sticky tags and dates are considered by this function.
# $fields is a comma separated string.
sub getCurrentFields {
   my($keyname,$fields) = (@_);

   my $sticky = $sticky{$keyname};
   my($latest,$date);
   my $tag = "MAIN";
   if(defined $sticky) {
      if($sticky =~ /^D/) {
         $latest = str2time(substr($sticky,1));
         $date = substr($sticky,1);
      } elsif($sticky =~ /^T/) { 
         $tag = substr($sticky,1);
      } elsif(length($sticky)) {
         sockprint("E bad sticky \"$sticky\" on $keyname\n");
      };
   }
   my @fields = ();
   my $id = cvsdb::id($keyname);
   if($id != -1) {  # if the file exists in the database
      my @files = cvsdb::GetFiles($fields,"dirent.file_id = $id","fname",$latest,$tag,$tag);
      @fields = @{$files[0]} if defined $files[0];
   }
   return ($id,     # The file id or -1 if not existent in database
           $latest, # A unix timestamp for the sticky date of the file.  Otherwise undef. 
           $tag,    # The sticky tag of the file (which may be a branchname).  Otherwise undef.
           $date,   # The CVS datestring format for the sticky date of the file.  Otherwise undef.
           @fields);  # (can also be undef)
}

# Sadly each file can have different sticky stuff, so we have to query each file individually.
# We are going to differ from CVS when no Entry is present, we inherit sticky from the dir and report it.
sub doStatus {
   my @args = map { defined $rev{"$rdir/$_"} ? () : "$rdir/$_" } @args;  # keep args if the don't have an Entry
   foreach my $keyname (sort keys(%rev),@args) {  # for each Entry line sent by the client
      $keyname =~ /^(.*?)\/?([^\/]*)$/; # match after last slash if present
      my($rdir,$filename,$complain) = ($1,$2,"");
      if(! defined $rev{$keyname}) {  # if the client did not send an Entry line
         $complain = "no file ";
         $sticky{$keyname} = $stickyDir{$rdir};   # cheat 
      };
      my($id,$latest,$tag,$date,$repoRev) = getCurrentFields($keyname,"revision.revision");
      $tag = "(none)" if ! defined $tag;
      $date = "(none)" if ! defined $date;
      my $revision = "Unknown file";
      my $status = "Unkown";
      my $working = $rev{$keyname} || "No entry for $filename";
      if($id != -1) {  # if the file does not exist in the database
         $revision = "$rev{$keyname}  $config::root$keyname  id=$id";
         if(! defined $rev{$keyname}) {
            $status = "Need Checkout";
         } elsif($repoRev eq $rev{$keyname}) {
            if(defined $text{$keyname}) {
               $status = "Locally Modified";
            } else {
               $status = "Up-to-date";
            };
         } else {  # else it is based on a old revision
            if(defined $text{$keyname}) {
               $status = "Needs Merge";
            } else {
               $status = "Needs Patch";
            };
         };
      }
      my $options = "(none)";
      $options = $kopt{$keyname} if defined $kopt{$keyname};
      sockprint("M ===================================================================\n" . 
                "M File: $complain$filename        Status: $status\n".
                "M\n".
                "M    Working revision: $working\n".
                "M    Repository revision: $revision\n".
                "M    Sticky Tag: $tag\n".
                "M    Sticky Date: $date\n".
                "M    Sticky Options: $options\n".
                "M\n");
   };
   sockprint("ok\n");
}

#   This is called when we need to fabricate a new file with a dead trunk revision,
#   so we can add a branched revision to the repository.
#   It will create a comment if need be.
sub createStillborn {
   my($comment_id_ref,$branchname,$keyname,$path,$filename) = (@_);
   sockprint("M RCS file: $config::root$keyname,v\n");
   if(! defined $$comment_id_ref) {  # if we don't yet have a comment for the stillborn
      db::SendSQL("INSERT comment SET descr = ". db::SqlQuote("initially added only for a branch (probably $branchname)") . 
                  ", author = " . db::SqlQuote($username) .", timestamp = " . $db::now);  
      db::SendSQL("SELECT LAST_INSERT_ID()");
      $$comment_id_ref = db::FetchOneColumn();
   };
   db::SendSQL("INSERT dirent SET fname = ".db::SqlQuote($filename).", parent = ".cvsdb::id($path).", type = 'file'");
   db::SendSQL("SELECT LAST_INSERT_ID()");
   my $id = db::FetchOneColumn();
   db::SendSQL("INSERT btime SET timestamp = NULL, file_id = $id, branchname='MAIN', branchnum=''");
   db::SendSQL("INSERT btime SET timestamp = NULL, file_id = $id, branchname = ".
               db::SqlQuote($branchname) . ", branchnum = '1.1.2'");   # tagging
   my $branchnum = "1.1.2";
   db::SendSQL("INSERT revision SET comment_id = $$comment_id_ref, file_id = $id, revision = '1.1'".
            ", state = 'dead', branchnum = " . db::SqlQuote($branchnum) . ", branches = ''".
            ", timestamp = $db::now, text = '', length = 0, md5 = '1B2M2Y8AsgTpgAmY7PhCfg', numAdded=0, numRemoved=0");
   db::SendSQL("SELECT LAST_INSERT_ID()");
   my $rev_id = db::FetchOneColumn();
   sockprint("M done\n");
   return($id,$branchnum,"1.1",$rev_id);
}

#  Sadly, the cvs client only tells us about the files it thinks are changed, so we cannot check
#  for out of date on the other ones.
#  The client presents each file with an Entry and as an Arguement, so we will only watch the Entries.
#  The files listed in Arguement bear no resemblance to what the user may have types (or not).
#
#  Our strategy is to first check that every file is up to date,
#  then create the comment in the database,
#  then compose all the SQL (so later we can delay locking) while composing response to client,
#  then send the SQL, and unlock tables,
#  and finally send response to client.
#  One exception is creating dirents and trunk-revisions for new adds on a branch (done during SQL compose time).
#
sub doCheckin {
   my($message,$force) = ("");
   while(my $flag = shift @args) {
      if($flag eq "-m") {
         $message = shift @args;
      } elsif($flag eq "-f") {
         $force++;
      } else {
         #unshift(@args,$flag);
         last;
      } 
   }   
   # Brain dead locking for now.
   db::SendSQL("LOCK TABLES dirent WRITE, revision WRITE, btime WRITE, comment WRITE, label READ, btime AS bt2 WRITE");
   # Check that everything is up to date
   my $fatal;
   my $acl = pack("nn",0,0); # deny everyone
   my(%rev_id,%newstate);
   foreach my $keyname (sort keys %rev) {  # for every Entry line sent by the client
      $newstate{$keyname} = ($rev{$keyname} =~ s/^-//) ? "dead" : "Exp";  # remove leading dash indicating removed
      (my $id,undef,undef,undef,my $repoRev,my $rev_id,$acl,my $owner) = 
          getCurrentFields($keyname,"revision.revision, revision.rev_id, latest.acl, latest.owner");
      if($id == -1 && ($rev{$keyname} eq "0" || $rev{$keyname} eq "import")) {  # if not in database and is a new-add or import
         # This was a new add and is not already in the database.  no problem.
      } elsif(! defined $rev_id) {
         $fatal++;
         sockprint("E dcvs server: I don't know the up-to-date revision of $keyname\n");
      } elsif($repoRev ne $rev{$keyname}) {
         $fatal++;
         sockprint("E dcvs server: $keyname is not up to date. (your $rev{$keyname} and our $repoRev)\n");
      } elsif(defined $conflict{$keyname} && $conflict{$keyname} eq "+=") {
         $fatal++;
         sockprint("E dcvs server: file '$keyname' had a conflict and has not been modified\n");
      } elsif(!session::isPrivilegedTo($acl,0x0010,$owner)) {  # if not privileged to commit 
         $fatal++;
         sockprint("E dcvs server: insufficient privileges to commit to $keyname\n") 
      }
      $rev_id{$keyname} = $rev_id;
   }
   if($fatal) {
      sockprint("E dcvs server: please update, test, and re-commit\n");
      db::SendSQL("UNLOCK TABLES");
      sockprint("ok\n");
      return;
   } 
   # Since everything is up to date, figure out the diffs and compose all the SQL
   db::SendSQL("INSERT comment SET descr = ". db::SqlQuote($message) . ", author = " . db::SqlQuote($username) .
               ", timestamp = " . db::getDbNowTime() .", acl = ".db::SqlQuote($acl));  # delete later if we need to
   db::SendSQL("SELECT LAST_INSERT_ID()");
   my $comment_id = db::FetchOneColumn();
   my $stillborn_comment_id;  #  in case we create a branched file for one without a trunk 
   my @sql = ();
   my $feedback = "";
   foreach my $keyname (sort keys %rev) {  # for every Entry line sent by the client
      my $id = cvsdb::id($keyname);
      $keyname =~ /(.*?)\/?([^\/]+)$/;
      my ($path,$filename) = ($1,$2);
      my $oldrev = $rev{$keyname};
      my $newadd = ($oldrev !~ /\./);  # typically the $oldrev of new adds is "0"
      my @lines2 = cvsdb::GetText($id,$oldrev) if ! $newadd;
      $text{$keyname} = join("",@lines2) if $newstate{$keyname} eq "dead";  # shameless
      if(! defined $text{$keyname}) {
         sockprint("E dcvs server: Error checkin $keyname but no Modified statement.\nerror\n");
         return;  # abort checkin
      }
      my $md5 = md5_base64($text{$keyname});
      my @lines = split(/^/m,$text{$keyname});
      my $newrev;  
      $sticky{$keyname} = "" if ! defined $sticky{$keyname};  # make sure sticky is defined
      if($sticky{$keyname} =~ /^T/) {  # if we are on a branch
         my $branchname = substr($sticky{$keyname},1);
         my($branchnum);
         if($newadd) {  # if this is a newly added file  (then create the file and trunk revision now)
            ($id,$branchnum,$oldrev,$rev_id{$keyname}) = createStillborn(\$stillborn_comment_id,$branchname,$keyname,$path,$filename);
            #@lines = ();
         } else {  # else this is an existing file (not new add)
            db::SendSQL("SELECT branchnum FROM btime WHERE timestamp IS NULL AND file_id = $id AND branchname = ".
                        db::SqlQuote($branchname));
            $branchnum = db::FetchOneColumn();
         }
         if(substr($oldrev,0,length($branchnum)) eq $branchnum) { # if $oldrev has $branchnum prefix (i.e. $branchnum is shorter)
            $oldrev =~ /\.([^\.]+)$/;  # match last digit
            $newrev = "$branchnum." . ($1 + 1);
            push(@sql,"UPDATE revision SET next = ". db::SqlQuote($newrev) ." WHERE rev_id = $rev_id{$keyname}");
         } else { # else this is the first off-trunk revision for this branch
            $newrev = "$branchnum.1";
            # We need to list the new revision in the "branches" of the trunk revision, even though dcvs doesn't use it.
            db::SendSQL("SELECT branches FROM revision WHERE file_id = $id AND revision = ".db::SqlQuote($oldrev));
            my($branches) = db::FetchSQLData();
            push(@sql,"UPDATE revision SET branches = ". db::SqlQuote("$branches $newrev") ." WHERE rev_id = $rev_id{$keyname}");
         }
         my($diff,$numAdded,$numRemoved) = rcsdiff::rcsdiff(\@lines2,\@lines);
         push(@sql,"INSERT revision SET comment_id = $comment_id, file_id = $id, revision = ". db::SqlQuote($newrev) . 
                  ", state = '$newstate{$keyname}', branchnum = " . db::SqlQuote($branchnum) .
                  ", timestamp = $db::now, text = ". db::SqlQuote($diff). ", md5 = ".db::SqlQuote($md5) .
                  ", length = ". length($text{$keyname}) .", added=$numAdded, removed=$numRemoved");
      } else {  # else we are on the main trunk
         my($numAdded,$numRemoved) = (0,0);
         if($newadd) {  # if this is a newly added file  ( typically $oldrev of 0 )
            #sockprint("M RCS file: $config::root$keyname,v\n");
            $newrev = "1.1";
            $oldrev = "";
            db::SendSQL("INSERT dirent SET fname = ".db::SqlQuote($filename).", parent = ".cvsdb::id($path).
                        ", type='file', head='1.1', mode = ".db::SqlQuote($mode{$keyname}) . " numAdded=0,numRemoved=0");
            db::SendSQL("SELECT LAST_INSERT_ID()");
            $id = db::FetchOneColumn();
            db::SendSQL("INSERT btime SET timestamp = NULL, file_id = $id, branchname='MAIN', branchnum=''");
            #sockprint("M done\n");
         } else {  # else this is a derivative of the trunk head
            $oldrev =~ /(.*?)\.([^\.]+)$/;  # match last digit
            $newrev = "$1." . ($2 + 1);
            (my $diff,$numAdded,$numRemoved) = rcsdiff::rcsdiff(\@lines,\@lines2);
            # Only the head revision is an actual text, the other texts are actually rcsdiff's
            push(@sql,"UPDATE revision SET text = ". db::SqlQuote($diff) ." WHERE rev_id = $rev_id{$keyname}");
            push(@sql,"UPDATE dirent SET head = ".db::SqlQuote($newrev).", mode = ".db::SqlQuote($mode{$keyname})
                      ." WHERE file_id = $id");
         }
         push(@sql,"INSERT revision SET comment_id = $comment_id, file_id = $id, revision = ". db::SqlQuote($newrev) . 
                     ", state = '$newstate{$keyname}', branchnum = '', next = ". db::SqlQuote($oldrev) .
                     ", timestamp = $db::now, text = ". db::SqlQuote($text{$keyname}). ", md5 = ".db::SqlQuote($md5) .
                     ", length = ". length($text{$keyname}) .", added=$numAdded, removed=$numRemoved");
      }; 
      if($rev{$keyname} eq "import") {  # if this was an import
         $feedback .= "M N $keyname\n";
      } elsif($newstate{$keyname} eq "dead") {  # if this commit removes the file
         $feedback .= "M Removing $filename;\n".
                      "M $config::root$keyname,v  <-- $filename\n".
                      "M new revision: delete; previous revision: $oldrev\n".
                      "Remove-entry ./$rdir\n".
                      "$config::root$keyname\n".
                      "M done\n";
      } else {  # else we did not remove it
         $feedback .= "M Checking in $filename;\n".
                      "M $config::root$keyname,v  <-- $filename\n".
                      "M new revision: $newrev; previous revision: $oldrev\n".
                      "Mode $mode{$keyname}\n".
                      "Checked-in ./$rdir\n".
                      "$config::root$keyname\n".
                      "/$filename/$newrev///$sticky{$keyname}\n".
                      "M done\n";
      };
   };  # end for every file
   # Now that all the SQL is composed we are ready to change the database.
   if($#sql > -1) {  # if we have any actual changes to introduce
      while(my $sql = shift @sql) {
         db::SendSQL($sql);
      };
   } else {   # else we created no new revisions, so remove the comment
      db::SendSQL("DELETE comment WHERE comment_id = $comment_id");  # because no revisions
   }
   db::SendSQL("UNLOCK TABLES");
   # Wow, that was easy, let's brag to the user now.
   sockprint($feedback);
   sockprint("ok\n");
}

# Here is the plan:   Fudge things up so it looks just like a commit, then call doCheckin.
sub doImport {
   my @args2;   # here are the fudged args we will be using.
   my($branchnum,$path,$vendor,$release) = ("1.1.1",undef,"vendor","imported");
   while(my $flag = shift @args) {
      if($flag eq "-b") {
         $branchnum = shift @args; 
      } elsif($flag eq "-m") {
         push @args2, $flag, shift @args;  # put the -m flag and the message arg
      } else {
         $path = $flag;
         ($vendor,$release) = @args;
         last;
      }
   }
   @args = @args2;
   foreach my $keyname (sort keys %text) {  # for each file to import
      $rev{$keyname} = "import";
      #$kopt{$keyname} = ;
      $sticky{$keyname} = "";
      # We need to create directories
      $keyname =~ /^(.*)\//s;  # get the directory
      cvsdb::id($1,1);  # make directory flag
   }
   return &doCheckin;
}

# First we check everything is okay, then we send the SQL, and finally we tell the user.
sub doTag
{
   my($branchflag);
   while(my $flag = shift @args) {
      if($flag eq "-b") {
         $branchflag++;
      } else {
         unshift(@args,$flag);
         last;
      } 
   }   
   my $branchname = shift @args;
   if(! defined $branchname || $branchname eq "") {
      sockprint("E dcvs server: missing tag name Arguement\nerror\n");
      return;
   }
   my($feedback,$fatal,@sql) = ("");
   foreach my $keyname (sort keys %rev) {  # for each Entry line sent by the client
      # we have to know the file_id
      my $id = cvsdb::id($keyname);
      if(! defined $id) {
         sockprint("E dcvs server: I don't know file $config::root$keyname\n");
         $fatal++;
      }
      # check for an existing branch under that name
      db::SendSQL("SELECT branchnum,acl,owner FROM btime WHERE file_id = $id AND branchname = ".
                  db::SqlQuote($branchname)." AND timestamp IS NULL");      
      my($branchnum,$acl,$owner) = db::FetchSQLData();
      if(defined $branchnum) {
         sockprint("E dcvs server: file $config::root$keyname already has a branch $branchname ($branchnum)\n");
         $fatal++;
         next;
      };
      my $privflag = $branchflag ? 0x0400 : 0x4000;  # create-b (branch) or create-l (label)
      if(session::isPrivilegedTo($acl,$privflag,$owner)) {
         sockprint("E dcvs server: file $config::root$keyname no permission to ".
                   ($branchflag ? "branch" : "label") . "\n");
         $fatal++;
         next;
      };
      # check for an existing tag under that name
      db::SendSQL("SELECT revision FROM label WHERE file_id = $id AND label = ".db::SqlQuote($branchname));      
      my($revision) = db::FetchSQLData();
      if(defined $revision) {
         sockprint("E dcvs server: file $config::root$keyname already has a tag $branchname ($revision)\n");
         $fatal++;
         next;
      };
      # find the time of the revision so we can make the branch entries
      db::SendSQL("SELECT timestamp FROM revision WHERE file_id = $id AND revision = ".db::SqlQuote($rev{$keyname}));      
      my($timestamp) = db::FetchSQLData();
      if(! defined $timestamp) {
         sockprint("E dcvs server: file $config::root$keyname does not have a revision ($revision) which you request.\n");
         $fatal++;
         next;
      };
      if($branchflag) {  # if this is a branch and not a tag
         # find a unique branch num for this file
         $branchnum = "$rev{$keyname}.";
         my $highest = 0;
         db::SendSQL("SELECT branchnum FROM btime WHERE file_id = $id AND branchnum LIKE ".db::SqlQuote("$branchnum%")." AND timestamp IS NULL");
         while(my $badnum = db::FetchOneColumn()) {
            die if(substr($badnum,0,length($branchnum)) ne $branchnum);  # if has wrong prefix 
            next if substr($badnum,length($branchnum)) =~ /\./; # if this was a branch of a branch at our branchpoint
            substr($badnum,length($branchnum)) =~ /(\d+)/;
            die if($1 == $highest);
            $highest = $1 if $1 > $highest;
         }
         $branchnum .= $highest + 2;
         # insert the branch table instructions
         push(@sql,"INSERT btime SET file_id = $id, branchname = ".db::SqlQuote($branchname).
                   ", branchnum = ".db::SqlQuote($branchnum).", timestamp = NULL");
         my @branchbranch = split(/\./,$branchnum);
         pop @branchbranch; pop @branchbranch;  # remove last number  (e.g. A.B.C.D.E becomes A.B.C)
         while(1) {  # while there more branches before the trunk
            my $branchbranch = join(".",@branchbranch);
            $branchbranch = "" if $#branchbranch <= 1;  # if X then we are on the trunk and use ""
            push(@sql,"INSERT btime SET file_id = $id, branchname = ".db::SqlQuote($branchname).
                      ", branchnum = ".db::SqlQuote($branchbranch).", timestamp = $timestamp") if $timestamp;
            last if $branchbranch eq "";
            pop(@branchbranch); 
            my $branchpoint = join(".",@branchbranch);
            db::SendSQL("SELECT timestamp FROM revision WHERE file_id = $id AND revision = ".db::SqlQuote($branchpoint));      
            $timestamp = db::FetchSQLData();
            sockprint("E dcvs server: database missing revision $branchpoint of $keyname\n") if ! $timestamp;  # database inconsistency
            pop(@branchbranch);
         }
      } else {  # else this is a tag, not a branch
         push(@sql,"INSERT label SET file_id = $id, label = ".db::SqlQuote($branchname).
                   ", revision = ".db::SqlQuote($rev{$keyname}));
      };
      $feedback .= "M T $keyname\n";
   };
   if($fatal) {
      sockprint("E dcvs server: Please fix this problem and retry. (database unchanged)\nok\n");
      return;
   };
   if($#sql == -1) {  # if there are no SQL statements
      sockprint("E dcvs server: nothing to do?\nok\n");
      return;
   };
   foreach my $sql (@sql) {
      db::SendSQL($sql);
   }
   sockprint($feedback);
   sockprint("E dcvs server: Don't forget to 'update -r $branchname'\n".
             "E dcvs server: if you want to check in on the new branch.\n") if $branchflag;
   sockprint("ok\n");
}

my %commandFunc = ( add => \&doAdd,
                    Argument => \&doArgument,
                    Argumentx => \&doArgumentx,
                    co => \&doCheckoutOrUpdate,
                    "Checkin-time" => \&doCheckinTime,
                    ci => \&doCheckin,
                    diff => \&doDiff,
                    Directory => \&doDirectory,
                    Entry => \&doEntry,
                    import => \&doImport,
                    "Is-modified" => \&doIsModified,
                    Kopt => \&doKopt,
                    log => \&doLog,
                    Modified => \&doModified,
                    noop => \&doNoop,
                    remove => \&doRemove,
                    Root => \&doRoot,
                    Questionable => \&doQuestionable,
                    status => \&doStatus,
                    Sticky => \&doSticky,
                    "Static-directory" => \&doStaticDirectory,
                    tag => \&doTag,
                    Unchanged => \&doUnchanged,
                    update => \&doCheckoutOrUpdate,
                    UseUnchanged => \&doUseUnchanged,
                    "Valid-responses" => \&doValidResponses,
                    "valid-requests" => \&doValidRequests,
                    "version" => \&doVersion );

sub doClientSession
{
   doAuthenticate() || return;
   while(defined(my $line = $sock->getline())) {
      print(">>>$line") if $nofork;
      chomp($line);
      my($command,$tail) = split(/\s+/,$line,2);
      #$command =~ tr/A-Z/a-z/;  # make lower case
      if(exists($commandFunc{$command})) {
         my $funcref = $commandFunc{$command};
         &$funcref($tail,$command);
      } else {
         push(@args,$line);
         print("arg? $line\n");
      }
   }
}

sub REAPER {
   my $waitedpid = wait;
   $SIG{CHLD} = \&REAPER;  # loathe sysV
   print("reaped $waitedpid\n");
}
 
$SIG{CHLD} = \&REAPER if(! $nofork);
 
my $server = IO::Socket::INET->new( Proto     => 'tcp',
                                    LocalPort => $port,
                                    LocalAddr => $localaddr,
                                    Listen    => SOMAXCONN,
                                    Reuse     => 1);
$server || die "can't open socket: $!";
 

if(!$nofork) {
   fork() and exit;  # daemonize (parent dies)
};
close(STDIN);
while(($sock,$peer) = $server->accept()) {
    $sock->autoflush(1);
    if($nofork) {
       doClientSession($sock);
       print("connection closed\n");
       resetStateForNofork();
    } else {  # else we for each connection
       if(!defined(my $pid = fork())) { 
          print(STDERR "Cannot fork(): $!\n");
       } elsif($pid) {  # else this is the child
           print("accepted connection pid=$$\n");        
           doClientSession($sock);
           exit(0);
       };
    };
    close $sock;
}




