#!/usr/bin/perl -w
use strict;
#use diagnostics;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

use Getopt::Long;
my($verbose) = (1);
GetOptions ('verbose' => \$verbose,
            'quiet'   => sub { $verbose = 0 });

# Add the directory of the executable to the module search path.
BEGIN { unshift @INC, "/var/www/html/dcvs"; }

use cvsdb;

sub removeFile
{
   my($file_id) = @_;

   db::SendSQL("DELETE FROM dirent WHERE file_id = $file_id");
   db::SendSQL("DELETE FROM btime WHERE file_id = $file_id");
   db::SendSQL("DELETE FROM label WHERE file_id = $file_id");
   db::SendSQL("DELETE FROM revision WHERE file_id = $file_id");
   db::SendSQL("DELETE FROM descendent WHERE parent = $file_id");
   db::SendSQL("DELETE FROM descendent WHERE child = $file_id");
}

sub removeAbandonedComments
{
   db::SendSQL("CREATE TEMPORARY TABLE orphans SELECT comment.comment_id FROM comment ".
               "LEFT JOIN revision USING (comment_id) ".
               "WHERE revision.comment_id IS NULL");
   db::SendSQL("SELECT comment_id FROM orphans");
   my @ids;
   while(my $id = db::FetchOneColumn()) {
      push(@ids,$id);
   }
   db::SendSQL("DROP TABLE orphans");
   for my $id (@ids) {
      db::SendSQL("DELETE FROM comment WHERE comment_id = $id");
   }
}

sub remove
{
   my($file_id,$onlyChildrenFlag) = @_;

   return if $file_id == -1;
   db::SendSQL("SELECT file_id,type FROM dirent WHERE parent = $file_id");
   my(@files,@dirs);
   while(my($file_id,$type) = db::FetchSQLData()) {
       if($type eq "dir") {
          push(@files,$file_id);
       } else {
          push(@dirs,$file_id);
       }
   }
   for my $id (@files) {
      removeFile($id);
   }
   for my $id (@dirs) {
      remove($id);
   }
   removeFile($file_id) if ! $onlyChildrenFlag;
}

if($#ARGV <= -1) {
   print("usage: rm-dcvs.pl filename ...\n");
   print("Careful! This removes ALL revisions.");
   print("Trailing slash on directories, means the directory contents not the directory itself.\n");
   exit;
}

foreach my $argv (@ARGV) {
    my $id = cvsdb::id($argv);
    if($id == -1) {
       print("File $argv does not exist.\n");
       next;
    };
    my $onlyChildrenFlag = ($argv =~ /\/$/);  # ends with slash
    remove($id,$onlyChildrenFlag);
    removeAbandonedComments();
}


