#!/usr/bin/perl -w
use strict;
#use diagnostics;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

use Date::Format;

use Getopt::Long;
my($verbose) = (1);
GetOptions ('verbose' => \$verbose,
            'quiet'   => sub { $verbose = 0 });

# Add the directory of the executable to the module search path.
BEGIN { unshift @INC, "/var/www/html/dcvs"; }

use cvsdb;

#
#  Please see the rcsfile (5) man page for details on the fileformat
#

# Revisions on the MAIN trunk (one period) go first with 1.1 last.
# Revisions not on the MAIN trunk go next in reverse order.
sub deltacmp
{
   my $r1 = $$a[0];  # revision num is first element is list.  $a is list ref
   my $r2 = $$b[0];
   my $r1Trunk = ($r1 =~ /^[^\.]+\.[^\.]+$/);
   my $r2Trunk = ($r2 =~ /^[^\.]+\.[^\.]+$/);
   my $sign = $r1Trunk ? -1 : 1;
   return -1 if($r1Trunk && !$r2Trunk);
   return 1 if(!$r1Trunk && $r2Trunk);
   my @R1 = split(/\./,$r1);
   my @R2 = split(/\./,$r2);
   while(my $e1 = shift(@R1)) {
      my $e2 = shift(@R2);
      return $sign if !defined($e2);
      my $x = ($e1 <=> $e2);
      return ($sign * $x) if $x;
   }
   return -$sign if defined(shift(@R2));
   return 0;
}

# sort 2 elements or shorter in forward direction, longer ones backward
# The order of branches from the same branch point doesn't matter.
sub deltatextcmp
{
   my $r1 = $$a[0];  # revision num is first element is list.  $a is list ref
   my $r2 = $$b[0];
   my @R1 = split(/\./,$r1);
   my @R2 = split(/\./,$r2);
   my $length = 0;
   while(1) {
      $length++;
      my $e1 = shift(@R1);
      my $e2 = shift(@R2);
      return -1 if !defined($e1);
      return 1 if !defined($e2);
      my $x = ($e1 <=> $e2);
      return ($length>3 ?$x:-$x) if $x;
   }
}

sub saveFile
{
   my($rcsname,$file_id,$comment,$head,$expand) = @_;

#$expand ||= "";
#print("saveFile(rcsname=$rcsname file_id=$file_id comment=$comment head=$head expand=$expand)\n");
   if(!open(RCSFILE,">$rcsname")) {
      print("Could not create file $rcsname\n");
      close(RCSFILE);
      return;
   }
   print(RCSFILE "head $head;\naccess;\nsymbols");
   db::SendSQL("SELECT label,revision FROM label WHERE file_id = $file_id");
   while(my($label,$revision) = db::FetchSQLData()) {
      print(RCSFILE "\n\t$label:$revision");
   };
   db::SendSQL("SELECT branchname,branchnum FROM btime WHERE file_id = $file_id AND timestamp IS NULL");
   while(my($branchname,$branchnum) = db::FetchSQLData()) {
      next if $branchname eq "MAIN";
      $branchnum =~ s/(\d+)$/0.$1/;  #   convert X.Y.Z into X.Y.0.Z
      print(RCSFILE "\n\t$branchname:$branchnum");
   };
   $comment =~ s/@/\@\@/g;
   print(RCSFILE ";\nlocks; strict;\ncomment \@$comment\@;\n" );
   print(RCSFILE "expand \@$expand\@;\n") if $expand;
   print(RCSFILE "\n\n");
   db::SendSQL("SELECT revision,revision.timestamp,author,state,branches,next,md5 FROM revision ".
               "LEFT JOIN comment USING (comment_id) ".
               "WHERE file_id = $file_id");
   my @rowrefs = db::FetchListOfLists();
   @rowrefs = sort deltacmp @rowrefs;
   while(my $rowref  = shift(@rowrefs)) {
      my($revision,$timestamp,$author,$state,$branches,$next,$md5) = @$rowref;
      my $date = ::time2str("%Y.%m.%d.%H.%M.%S",$timestamp,"UTC");
      $date =~ s/^(\d{4}\.)(\d\.)/${1}0$2/;  # add missing month leading zero if needed
      print(RCSFILE "$revision\ndate\t$date;\tauthor $author;\tstate $state;\nbranches\t"); 
      $branches =~ s/\s+/\n\t/;
      $next ||= "";  # silence undefined warnings
      print(RCSFILE " $branches;\nnext\t$next;\tmd5 $md5;\n\n");
   };
   print(RCSFILE "desc\n\@\@\n\n");
   db::SendSQL("SELECT revision,descr,text FROM revision ".
               "LEFT JOIN comment USING (comment_id) WHERE file_id = $file_id");
   @rowrefs = db::FetchListOfLists();
   @rowrefs = sort deltatextcmp @rowrefs;
   while(my $rowref  = shift(@rowrefs)) {
      my($revision,$descr,$text) = @$rowref;
      $descr =~ s/@/\@\@/g;
      $text =~ s/@/\@\@/g;
      print(RCSFILE "$revision\nlog\n\@$descr\@\ntext\n\@$text\@\n\n");
   };
   close(RCSFILE);
}

sub saveRcsRecurse {
   my ($diskPath,$dbPath,$where) = @_;

print("saveRcsRecurse(diskPath=$diskPath dbPath=$dbPath where=$where)\n");
   my @files = cvsdb::GetFiles("dirent.file_id,fname,type,state,comment,revision,expand",$where,"fname",undef,"MAIN");
   while(my $filelistref = shift(@files)) {
      my($file_id,$fname,$type,$state,$comment,$head,$expand) = @$filelistref;
      if($type eq "dir") {
          if(! -d "$diskPath$fname" && !mkdir("$diskPath$fname")) {
             print("Could not create directory $diskPath$fname\n");
          } elsif(! -d "$diskPath$fname/Attic" && !mkdir("$diskPath$fname/Attic")) {
             print("Could not create directory $diskPath$fname/Attic\n");
          } else {
             saveRcsRecurse("$diskPath$fname/","$dbPath$fname/","dirent.parent = $file_id");
          };
      } elsif($type eq "file") {
          if($state eq "dead") {
             saveFile("${diskPath}Attic/$fname,v",$file_id,$comment,$head,$expand);
          } else {
             saveFile("$diskPath$fname,v",$file_id,$comment,$head,$expand);
          };
      } else {
          print("$dbPath$fname could not export type $type.\n");
      };
   }   
}

# Trailing slash on directories is relevant.
# diskPath must specify a directory
sub saveRcs {
   my ($diskPath,$dbPath) = @_;

   $|++;  # disable output buffering
   $dbPath = "/" if ! defined($dbPath);
   my $dbPathId = cvsdb::id($dbPath);
   return if $dbPathId == -1;
   $diskPath .= "/" unless $diskPath =~ /\/$/;  # unless ends in slash
   if(! -d $diskPath) {
       print("Local filesystem directory $diskPath does not exits.\n");
       return;
   }
   return if ! -d $diskPath;
   if($dbPath =~ /\/$/) {  # if ends in a slash
      if(! -d "${diskPath}Attic" && ! mkdir("${diskPath}Attic")) {
         print("Could not create directory $diskPath/Attic.\n");
         return;
      }
      saveRcsRecurse($diskPath,$dbPath,"dirent.parent = $dbPathId");
   } else {
      saveRcsRecurse($diskPath,$dbPath,"dirent.file_id = $dbPathId");
   };
   $|--;  # possibly re-enable output buffering
}


if($#ARGV < 1) {  # if one command line parameter or fewer
   print("usage: savercs.pl diskPath dbPath ...\n");
   print("(Trailing slash on dbPath is significant.)\n");
   exit 1;
}
my $diskPath = shift @ARGV;
foreach my $dbPath (@ARGV) {
   $dbPath =~ s/\/+$//;  # remove trailing slashes
   saveRcs($diskPath,$dbPath);
}


1;

