#!/usr/bin/perl -w
use strict;

use useradmin;
use db;
use config;

use Getopt::Long;
my($createFlag,$forceFlag);
GetOptions ('create' => \$createFlag,
            'force'  => \$forceFlag );

if($#ARGV != 1) {  # if not two remaining command line parameters
   print("usage: setdcvspasswd.pl [ options ] username password\n");
   print("    --create      Create account which does not exist.\n");
   print("    --force       Force allow bad passwords.\n");
   print("The password \"os\" is special and will instead check with the operating system\n");
   print("at every login attempt.\n");
   exit;
}
my($username,$password) = @ARGV;

if(-u $0 || -g $0) {  # if executable is SUID for either user or group
   print(STDERR "This script must not be SUID.\n");
   print(STDERR "It does not require a password.\n");
   exit -6;
}

db::SendSQL("SELECT username FROM user WHERE username = ".db::SqlQuote($username));
my $exists = db::FetchSQLData();
if($createFlag) {
   if($exists) {
      print(STDERR "Username '$username' already exists.\n"); 
      exit -4;
   }
} else {  # no create
   if(!$exists) {
      print(STDERR "Username '$username' does not exist.  Use --create flag if needed.\n"); 
      exit -5;
   }
}

if($password eq "os") {
   useradmin::set_user($username,undef,undef,undef,undef,undef,"os");
} else {  # else use actual password
   if(length $password < $config::minPasswdLen && ! $forceFlag) {
      print(STDERR "Yuck! Password is too short (less than $config::minPasswdLen chars).  Use --force if need be.\n");
      exit -6;
   }
   if($password =~ /^[a-zA-Z]+$/ && ! $forceFlag) {  # if all alpha
      print(STDERR "Yuck! Password is all alpha.  Use --force if need be.\n");
      exit -6;
   }
   useradmin::set_user($username,$password);
};




