#!/usr/bin/perl -w
use strict;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

BEGIN { push @INC, "/var/www/html/dcvs" };

#
#  This program is a learning aid for the perl diff::diff() function.
#  It will present the diff of two files in a perl data structure like diff::diff() creates.
#

use diff;
use Data::Dumper;

if($#ARGV != 1) {   # if not two command line args
   print("usage: perldiff file1 file2\n");
   exit -1;
}
my($filename1,$filename2) = @ARGV;

open(FD,"<$filename1") or die("could not open $filename1\n$!");
read(FD,my $file1,1000000);
close(FD);
open(FD,"<$filename2") or die("could not open $filename2\n$!");
read(FD,my $file2,1000000);
close(FD);

my @list1 = split(/^/m,$file1);
my @list2 = split(/^/m,$file2);
my $changes = diff::diff(\@list1,\@list2);
my $dump = Dumper($changes);

# Collapse the excessive number of lines
my $openBraces = 0;
foreach my $dumpline (split(/^/m,$dump)) { # for each line of the dump
   $dumpline =~ s/^\s+/ / if $openBraces >= 2;  # collapse leading white space to one space
   $openBraces-- if($dumpline =~ /^\s*\]/);  # if line begins with open square brace
   $openBraces++ if($dumpline =~ /^\s*\[/);  # if line begins with open square brace
   chomp($dumpline) if ($openBraces >= 2);
   print $dumpline;
}

__END__



         a b c e h j l m n p
         b c d e f j k l m r s t

       Result:

        [
          [ [ '-', 0, 'a' ] ],

          [ [ '+', 2, 'd' ] ],

          [ [ '-', 4, 'h' ] ,
            [ '+', 4, 'f' ] ],

          [ [ '+', 6, 'k' ] ],

          [ [ '-', 8, 'n' ],
            [ '-', 9, 'p' ],
            [ '+', 9, 'r' ],
            [ '+', 10, 's' ],
            [ '+', 11, 't' ],
          ]
        ]



