#!/usr/bin/perl -w
use strict;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

# Add the parent directory of the executable to the module search path.
BEGIN { push(@INC,"/var/www/html/dcvs"); }  # library include path

use rcsdiff;
use cvsdb;
use Date::Parse;
use Date::Format;

#
#  We go through a RCS file looking for rcsdiffs.  When
#  we find one we use co to grab the versions and try to
#  compose our own rcsdiff.  We then try to reconstruct
#  the derived version from both rcsdiffs.
#
#  This script makes more sense if you run it on an RCS
#  file not created by dcvs.
#
#  This (test) code is ugly and slow, please stay away.  
#
#  This code will puke on keyword substitution.
#

if($#ARGV == -1) {
   print("usage  rcsdifftest.pl rcsfilename ...\n");
   exit;
}

my $ansiBright = "\033[1m";
my $ansiBlue = "\033[34m";
my $ansiGreen = "\033[32m";
my $ansiRed = "\033[31m";
my $ansiNormal = "\033[0m";

sub showSideBySide($$) {
   my($lr1,$lr2) = @_;

   for(my $index = 0;$index < $#$lr1 && $index < $#$lr2;$index++) {
      my $line1 = $$lr1[$index];
      my $line2 = $$lr2[$index];
      $line1 = "" if ! defined $line1;
      $line2 = "" if ! defined $line2;
      $line1 =~ s/\t/ /g;
      $line2 =~ s/\t/ /g;
      $line1 = substr($line1,0,40);
      $line2 = substr($line2,0,40);
      chomp($line1);
      chomp($line2);
      $line1 .= " " x (40 - length($line1)) if length $line1 < 40;
      $line2 .= " " x (40 - length($line2)) if length $line2 < 40;
      if($line1 eq $line2) {
         print("$line1 ### $line2\n"); 
      } else {
         print("$ansiRed$line1 ### $line2$ansiNormal\n"); 
      };
   }
}


sub getRevText ($$) 
{
   my($rev,$filename) = @_;

   open(CO,"co -p$rev $filename 2>/dev/null|") or die;
   my $bytesRead = read(CO,my $text,1000000);
   defined $bytesRead or die("co -p$rev $filename\n ");
   close(CO);
   return $text;
}

sub compare($$$$) {
   my($filename,$late,$old,$rcsGood) = @_;

print("compare($filename  $late  $old   )\n");
   my $lateText = getRevText($late,$filename);
   my $oldText = getRevText($old,$filename);
   my @lateLines = split(/^/m,$lateText);
   my @oldLines = split(/^/m,$oldText);
   my $rcsSuspect = rcsdiff::rcsdiff(\@lateLines,\@oldLines);
   if($rcsGood ne $rcsSuspect) {
      #print(join("lateLines> ","",@lateLines),"\n\n");
      #print(join("oldLines> ","",@oldLines),"\n\n");
      #print(join("good> ","",split(/^/m,$rcsGood)),"\n\n");
      #print(join("susp> ","",split(/^/m,$rcsSuspect)),"\n\n");
      my @good = split(/^/m,$rcsGood);
      my @susp = split(/^/m,$rcsSuspect);
      #showSideBySide(\@good,\@susp);
      print("different diffs\n");
      #exit 1;
   }
   cvsdb::applyDiffs($rcsGood,\@lateLines);
   if($oldText ne join("",@lateLines)) {
      print(join("rcsGood> ","",split(/^/m,$rcsGood)),"\n\n");
      #print(join("GOOD> ","",split(/^/m,$oldText)),"\n\n");
      #print(join("BAD > ","",@lateLines),"\n\n");
      print("bad reconstruction (problem with applyDiffs)\n");
      my @good = split(/^/m,$oldText);
      showSideBySide(\@good,\@lateLines);
      exit 1;
   }
   @lateLines = split(/^/m,$lateText);
   cvsdb::applyDiffs($rcsSuspect,\@lateLines);
   if($oldText ne join("",@lateLines)) {
      print(join("GOOD> ","",split(/^/m,$oldText)),"\n\n");
      print(join("BAD > ","",@lateLines),"\n\n");
      print("bad reconstruction (problem with rcsdiff)\n");
      exit 1;
   };

   return($oldText,$rcsSuspect);
}





#
#  Please see the rcsfile (5) man page for details on the file format.
#

my $ungetParsedElem;
my $readBuffer = "";

sub UngetParsedElement
{
   defined($ungetParsedElem) and die("internal error");
   $ungetParsedElem = $_[0];
   undef;
}

#  This will get either a @-delimited string, a semicolin, or a token from the file.
#  Trans-call state is kept in $ungetParsedElem and $readBuffer
sub GetParsedElement
{
   my($fh) = @_;

   if(defined($ungetParsedElem)) {
      my $return = $ungetParsedElem;
      $ungetParsedElem = undef;
      return $return;
   }
   my $elem;
   while(1) {  # while we need to read more for our first token
      if($readBuffer =~ s/^\s*([^\s;]+)([\s;])/$2/s || $readBuffer =~ s/^\s*(;)//s) {
         $elem = $1;
         last;
      };
      my $newRead;
      if(read($fh,$newRead,100) == 0) {  # if running out of new data
         return undef if($readBuffer =~ /^\s+$/); 
         $readBuffer .= "\n";  # make sure we end file with white space
      } else {
         $readBuffer .= $newRead;
      };
   };
   return $elem if($elem =~ /^[^\@]/);  # doesn't begin with @
   $elem .= $readBuffer;
   while(1) { # while we need to read more to complete the string
      if($elem =~ s/^\@(([^\@]*@@)*[^\@]*)\@([^\@].*)/$1/s) {
         $readBuffer = $3;
         $elem =~ s/\@\@/\@/g;
         return $elem;
      };
      my $readLength = (length($elem) < 100) ? 100 : length($elem);
      my $newRead;
      if(read($fh,$newRead,$readLength) == 0) {  # if running out of new data
         die("file ended in middle of string");
      };
      $elem .= $newRead;
   }
}

sub readSection {
   my($fh,$hashref,$comment) = @_;

   my $first = GetParsedElement($fh);
   if($first =~ /^[\d\.]+$/) {
      $$hashref{'num'} = $first;
      $first = GetParsedElement($fh);
   }
   while(1) {
      my @elems;
      while(defined(my $elem = GetParsedElement($fh))) {
         last if($elem eq ";");
         push(@elems,$elem);
      };
      $$hashref{$first} = join(" ",@elems);
      $first = GetParsedElement($fh);
      if($first =~ /^[\d\.]+$/ || $first eq "desc") {
         UngetParsedElement($first);
         return $first;
      };
   }
}

# Will return undef if end of file is reached
sub readDeltaText {
   my($fh,$hashref,$comment) = @_;

   my $first = GetParsedElement($fh);
   return undef if(!defined($first));
   if($first !~ /^[\d\.]+$/) {
      die("bad num ($first) at start of deltatext");
   }
   $$hashref{'num'} = $first;
   my $log = GetParsedElement($fh);
   if($log ne "log") {
      die("expected 'log' found $log");
   };
   $log = GetParsedElement($fh);
   $$hashref{'log'} = $log;
   $first = GetParsedElement($fh);
   while(1) {
      last if($first eq "text");
      my @elems;
      while(defined(my $elem = GetParsedElement($fh))) {
         last if($elem eq ";");
         push(@elems,$elem);
      };
      $$hashref{$first} = join(" ",@elems);
      $first = GetParsedElement($fh);
   }
   $$hashref{'text'} = GetParsedElement($fh);
   return $$hashref{'num'};
}

# FIXME: what about revisions in the database not in the RCS file
sub loadRcsFile
{
   my($filenameWithPath) = @_;

   local(*RCS);
   if(!open(RCS,"<$filenameWithPath")) {
      print("<font color=red>open failed: $!</font><br>");
      return;
   };
   my %admin;
   for my $id qw(head branch access symbols locks comment expand strict) {
      $admin{$id} = "";
   }
   my $mode = "u=rw,g=rw,o=r";
   $mode = "u=rwx,g=rwx,o=rx" if -X RCS;
   my $nextline = readSection(*RCS,\%admin,"admin");
   my @revnums = ();  # store revisions in a list so we can sort them by revision number
   my %revs;
   while(1) {  # for every delta section, load it
      last if $nextline eq "desc";
      my %delta;
      for my $id qw(num date author state branches next) {
         $delta{$id} = "";
      }
      $nextline = readSection(*RCS,\%delta,"delta");
      $delta{'date'} =~ s/^(\d+)\.(\d+)\.(\d+)\.(\d+)\.(\d+)\.(\d+)$/$1-$2-$3 $4:$5:$6/; # to SQL format
      $delta{'timestamp'} = ::str2time($delta{'date'},"UTC");
      push(@revnums,$delta{'num'});
      $revs{$delta{'num'}} = \%delta;
   }
   my(%prevOfBranch,%diffParentOfChild);   # for reconstructing text for md5
   @revnums = sort cvsdb::rev_cmp @revnums ;  # so rev_id is ordered nicely 
   for my $revnum (@revnums) {  # for every delta section, store in database
      my $branchnum = "";
      if($revnum =~ /^(\d+\.\d+\..*)\.\d+$/) {  # if more than two decimals
         $branchnum = $1;
         if(defined $prevOfBranch{$branchnum}) {
            $diffParentOfChild{$revnum} = $prevOfBranch{$branchnum};
         } else {
            (my $branchpoint = $branchnum) =~ s/\.[^\.]+$//;  # remove last dot and num
            $diffParentOfChild{$revnum} = $branchpoint; 
         };
      } else { # else this is on the main trunk
         $diffParentOfChild{$prevOfBranch{""}} = $revnum if defined $prevOfBranch{""};  # main trunk goes opposite direction
      };
      $prevOfBranch{$branchnum} = $revnum;
      my $deltaref = $revs{$revnum};
   }
   my %numDiffChildren;  # so we no if we need to save the text when doing md5
   foreach my $parent (values %diffParentOfChild) { $numDiffChildren{$parent}++ };
   GetParsedElement(*RCS);  # desc
   GetParsedElement(*RCS);  # string
   my $firstDeltaTextFlag;
   my %savedText;
   while(1) {  # for every deltatext section
      my %deltatext;
      my $num = readDeltaText(*RCS,\%deltatext,"deltatext");
      last if ! defined($num);
      if(!defined($revs{$num})) {
         print("<font color=red>partial info for revision $num for ".
               cgi::htmlQuote($filenameWithPath) . "</font><br>");
         next;
      }
      my $textLinesRef;
      my $parent = $diffParentOfChild{$num};
      if(defined $parent) { # if this delta text is actually a delta
         if($numDiffChildren{$parent}-- <= 0) {    # if this is the last use of the saved text
            $textLinesRef = delete $savedText{$parent};
         } else {   # else this is not the last use so we need to make a copy
            my @lines = @{$savedText{$parent}};
            $textLinesRef = \@lines;
         } 
         if(! defined $textLinesRef) {
            die("<font color=red>missing text of $parent for $num of $filenameWithPath</font><br>\n");
            $textLinesRef = [];  # reference to empty list
         }
         compare($filenameWithPath,$parent,$num,$deltatext{'text'});
      } else {  # else the deltatext is an actual file not a delta because this revision is the head
         my @textLines = split(/^/m,$deltatext{'text'});  # The first deltatext is not a delta, but the actual text
         $textLinesRef = \@textLines;
      }
      $savedText{$num} = $textLinesRef if $numDiffChildren{$num};
      #my $md5 = ::md5_base64(join("",@$textLinesRef));
   }
   close(RCS);
}   

foreach my $argv (@ARGV) {   # for each command line parameter
   if(-d $argv) {
      print("skipping $argv because it is a directory.\n");
      next;
   }
   if($argv !~ /,v$/) {
      print("skipping $argv because it has a bad suffix.\n");
      next;
   }
   loadRcsFile($argv);
}

__END__

foreach my $argv (@ARGV) {   # for each command line parameter
   open(RLOG,"rlog $argv |") or die;
   my $prev;
   my @revisions;
   while(defined (my $line = <RLOG>)) {
      if(defined $prev && $prev =~ /^-----------/ && $line =~ /^revision\s+(\d+\.\d+)/) {
         push(@revisions,$1);
      }; 
      $prev = $line;
   }
   close(RLOG);
   my $lateRev = shift(@revisions);
   foreach my $oldRev (@revisions) {
      compare($lateRev,$oldRev,$argv);
      $lateRev = $oldRev;
   }
}


