#!/usr/bin/perl -w
use strict;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

# This is really ugly code.  Please don't look.

package unixdiff;  # diff format used in RCS (version control) files

use diff qw(diff);

# This routine requires two list references.
# The lists each have one line of file content per list element.
sub unixdiff
{
   my($f1,$f2,$prefix) = @_;

   my $diffs = diff($f1, $f2);
   my $answer = "";
   my $lineNumChange = 0;
   while(my $hunk = shift @$diffs) {
      # first process the removals
      my $linechange;
      my $remove = "";
      my $removeStart = undef;
      my $removeNumLines = 0;
      while($linechange = shift @$hunk) {
         my($action,$lnum,$text) = @$linechange;
         last if $action eq '+';
         $removeStart = $lnum if(!defined $removeStart);
         $removeNumLines++;
         $remove .= "$prefix< $text"; 
      };
      # then process the additions
      my $insert = "";
      my $insertStart = undef;
      my $insertNumLines = 0;
      while(defined $linechange) {
         my(undef,$lnum,$text) = @$linechange;
         $insertStart = $lnum if(!defined $insertStart);
         $insertNumLines++;
         $insert .= "$prefix> $text"; 
         $linechange = shift @$hunk;
      }
      if($removeNumLines) {
         if($insertNumLines) {
            my $r = $removeStart + 1;
            $r = "$r,".($r+$removeNumLines-1) if $removeNumLines > 1;
            my $i = $insertStart + 1;
            $i = "$i,".($i+$insertNumLines-1) if $insertNumLines > 1;
            $answer .= "$prefix${r}c$i\n$remove$prefix---\n$insert";
         } else {
            my $r = $removeStart + 1;
            $r = "$r,".($r+$removeNumLines-1) if $removeNumLines > 1;
            my $x = $removeStart - $lineNumChange;
            $answer .= "$prefix${r}d$x\n$remove";
         }
      } elsif($insertNumLines) {
         my $i = $insertStart + 1;
         $i = "$i,".($i+$insertNumLines-1) if $insertNumLines > 1;
         my $at = $insertStart + $lineNumChange;
         $answer .= "$prefix${at}a$i\n$insert";
      }
      $lineNumChange += - $insertNumLines + $removeNumLines;
   } 
   return $answer;
};

1;

__END__

From diff.pm....

Here is an example:  The diff of the following two sequences:
 
  a b c e h j l m n p
  b c d e f j k l m r s t
 
Result:
 
 [
   [ [ '-', 0, 'a' ] ],
 
   [ [ '+', 2, 'd' ] ],
 
   [ [ '-', 4, 'h' ] ,
     [ '+', 4, 'f' ] ],
 
   [ [ '+', 6, 'k' ] ],
 
   [ [ '-', 8, 'n' ],
     [ '-', 9, 'p' ],
     [ '+', 9, 'r' ],
     [ '+', 10, 's' ],
     [ '+', 11, 't' ],
   ]
 ]



From the diff info pages....

 
Detailed Description of Normal Format
-------------------------------------
 
   The normal output format consists of one or more hunks of
differences; each hunk shows one area where the files differ.  Normal
format hunks look like this:
 
     CHANGE-COMMAND
     < FROM-FILE-LINE
     < FROM-FILE-LINE...
     ---
     > TO-FILE-LINE
     > TO-FILE-LINE...
 
   There are three types of change commands.  Each consists of a line
number or comma-separated range of lines in the first file, a single
character indicating the kind of change to make, and a line number or
comma-separated range of lines in the second file.  All line numbers are
the original line numbers in each file.  The types of change commands
are:
 
`LaR'
     Add the lines in range R of the second file after line L of the
     first file.  For example, `8a12,15' means append lines 12-15 of
     file 2 after line 8 of file 1; or, if changing file 2 into file 1,
     delete lines 12-15 of file 2.
 
`FcT'
     Replace the lines in range F of the first file with lines in range
     T of the second file.  This is like a combined add and delete, but
     more compact.  For example, `5,7c8,10' means change lines 5-7 of
     file 1 to read as lines 8-10 of file 2; or, if changing file 2 into
     file 1, change lines 8-10 of file 2 to read as lines 5-7 of file 1.
 
`RdL'
     Delete the lines in range R from the first file; line L is where
     they would have appeared in the second file had they not been
     deleted.  For example, `5,7d3' means delete lines 5-7 of file 1;
     or, if changing file 2 into file 1, append lines 5-7 of file 1
     after line 3 of file 2.



