#!/usr/bin/perl -w
use strict;
#use diagnostics;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

package useradmin;
require db;
use Digest::MD5 qw(md5_base64);

sub set_user 
{
   my($username,$password,$realname,$email,$disabledtext,$groups,$pwtype) = @_;

   my $query = "";
   if(defined $pwtype && $pwtype =~ /os$/) {  # if password is os directed
      $query .= ", cryptpassword = 'os'";
   } elsif($password) {  # if password is defined and not empty
      db::SendSQL("SELECT RAND(), RAND(), RAND(), RAND()");
      my $data = time() . $$ . rand() . \@_ . $username . join($password,db::FetchSQLData());
      if(-r "/dev/urandom" && open(RAND,"</dev/urandom")) {
         read(RAND,my $randstuff,32);
         close(RAND);
         $data .= $randstuff;
      };
      my $salt = md5_base64($data);
      $salt = "md5" . md5_base64($salt.rand());
      my $hash = md5_base64($salt . $password);
      my $cryptpassword = $salt . $hash; 
      $query .= ", cryptpassword = ".db::SqlQuote($cryptpassword);
   }
   if(defined($realname)) {
      $query .= ", realname = ".db::SqlQuote($realname);
   }
   if(defined($email)) {
      $query .= ", email = ".db::SqlQuote($email);
   }
   if(defined($disabledtext)) {
      $query .= ", disabledtext = ".db::SqlQuote($disabledtext);
   }
   if(defined($groups)) {
      $query .= ", groups = ".db::SqlQuote($groups);
   }
   # We will not just use SQL REPLACE because that changes the userid.
   db::SendSQL("SELECT 1 FROM user WHERE username = ".db::SqlQuote($username));
   if(defined db::FetchOneColumn()) {
      $query =~ s/^,//;  # remove leading comma
      $query = "UPDATE user SET $query WHERE username = ".db::SqlQuote($username);
   } else {
      $query = "INSERT user SET username = ".db::SqlQuote($username). $query;
   }
   db::SendSQL($query);
}

# The input is a packed groups list like found in the user table.
sub groupList {
   my($groups) = (@_);
   my @grouplist;
   foreach my $groupid (unpack("n*",$groups)) {
      db::SendSQL("SELECT name FROM groups WHERE group_id = $groupid");
      my($name) = db::FetchSQLData();
      push(@grouplist,defined($name) ? $name : $groupid);
   }
   @grouplist = sort @grouplist;
   return \@grouplist;
}

# $groups is a space separated string of group names.
# The return value is undef on sucess, an error message otherwise.
sub set_groups() {
   my($username,$groups) = (@_);

   db::SendSQL("SELECT group_id, name FROM groups");
   my %groupIdOfName;
   while(1) {
      my($group_id,$name) = db::FetchSQLData();
      last if ! defined $group_id;
      $groupIdOfName{$name} = $group_id;
   }
   my(%id,@notfound);
   foreach my $name (split(/\s+/,$groups)) {
      if($name =~ /^\d+$/) {  # if name is an integer
         if($name > 1 && $name < 0x10000) {
            $id{$name} = 1;
         } else { 
            push(@notfound,$name); 
         }
      } elsif (defined $groupIdOfName{$name}) {
         $id{$groupIdOfName{$name}} = 1 if($groupIdOfName{$name} != 0); # zero is "everyone" is assumed
      } else {
         push(@notfound,$name); 
      }
   };
   if($#notfound >= 0) {
      return "Illegal group names: \"". join('"  "',@notfound) . '" ';
   };
   $groups = pack("n*",keys %id);
   db::SendSQL("UPDATE user SET groups = ".db::SqlQuote($groups)." WHERE username = ".db::SqlQuote($username));
   return undef;  # return "no error"
}


